"""
Package to gather all the tiles created in the application.

If you only have few widgets, a module is not necessary and you can simply use a tile.py file
In a big module with lot of custom tiles, it can make sense to split things in separate files for the sake of maintenance
If you use a module import all the functions here to only have 1 call to make.
"""

from .map_tile import *



"""
Creation of the Translator object associated with the application. 

Can be accessed via the foolowing code: ``from component.message import cm``
"""

from pathlib import Path

from sepal_ui.translator import Translator

# create a translator object
cm = Translator(Path(__file__).parent)



"""
All in one tile to reclassify GEE assets or local raster and create custom classifications.
"""

from pathlib import Path
from typing import Optional, Union

import ipyvuetify as v
from traitlets import link

from sepal_ui import aoi
from sepal_ui import sepalwidgets as sw
from sepal_ui.message import ms

from .reclassify_model import ReclassifyModel
from .reclassify_view import ReclassifyView
from .table_view import TableView

__all__ = ["ReclassifyTile"]


class ReclassifyTile(sw.Tile):

    result_dir: Union[Path, str] = ""
    "Directory to store the outputs (rasters, and csv_files)."

    model: Optional[ReclassifyModel] = None
    "the reclassify model to use with these inputs"

    reclassify_view: Optional[ReclassifyView] = None
    "a fully qualified ReclassifyView object"

    table_view: Optional[TableView] = None
    "a fully qualified TableView object"

    def __init__(
        self,
        results_dir: Union[Path, str] = Path.home() / "downloads",
        gee: bool = True,
        dst_class: Union[Path, str] = "",
        default_class: dict = {},
        aoi_model: Optional[aoi.AoiModel] = None,
        folder: str = "",
        **kwargs
    ) -> None:
        """
        All in one tile to reclassify GEE assets or local raster and create custom classifications.

        Args:
            results_dir: Directory to store the outputs (rasters, and csv_files). default to ~/downloads
            gee: Use GEE variant, to reclassify assets or local input. default True
            dst_class: the file to be used as destination classification. for app that require specific code system the file can be set prior and the user won't have the oportunity to change it
            default_class: the default classification system to use, need to point to existing sytem: {name: absolute_path}
            folder: the init GEE asset folder where the asset selector should start looking (debugging purpose)
        """
        # output directory
        self.results_dir = Path(results_dir)

        self.aoi_model = aoi_model
        # create the model
        self.model = ReclassifyModel(
            dst_dir=self.results_dir, gee=gee, aoi_model=self.aoi_model, folder=folder
        )

        # set the tabs elements
        self.reclassify_view = ReclassifyView(
            self.model,
            out_path=self.results_dir,
            gee=gee,
            default_class=default_class,
            aoi_model=aoi_model,
            save=True,
            folder=folder,
        ).nest_tile()

        self.table_view = TableView(out_path=self.results_dir).nest_tile()

        # create the tab
        tiles = [self.reclassify_view, self.table_view]

        self._tabs = v.Tabs(
            class_="mt-5",
            grow=True,
            # background_color='grey darken-4',
            v_model=0,
            children=[
                v.Tab(children=[t.title.children[0].children[0]], key=i)
                for i, t in enumerate(tiles)
            ],
        )

        self._content = v.TabsItems(
            v_model=0,
            children=[v.TabItem(children=[t], key=i) for i, t in enumerate(tiles)],
        )

        super().__init__(
            id_="reclassify_tile",
            title=ms.rec.tile.title,
            inputs=[self._tabs, self._content],
            **kwargs
        )

        # js interaction
        link((self._tabs, "v_model"), (self._content, "v_model"))



"""
Helpers to customize the display of sepal-ui widgets and maps.
"""

from pathlib import Path
from types import SimpleNamespace
from typing import Dict, Tuple

import ipyvuetify as v
from IPython.display import HTML, Javascript, display
from traitlets import Bool, HasTraits, observe

import sepal_ui.scripts.utils as su
from sepal_ui.conf import config

################################################################################
# access the folders where style information is stored (layers, widgets)
#

# the colors are set using tables as follow.
# 1 (True): dark theme
# 0 (false): light theme
JSON_DIR: Path = Path(__file__).parent / "json"
"The path to the json style folder"

CSS_DIR: Path = Path(__file__).parent / "css"
"The path to the css style folder"

JS_DIR: Path = Path(__file__).parent / "js"
"The path to the js style folder"

################################################################################
# define all the colors taht we want to use in the theme
#

DARK_THEME: Dict[str, str] = {
    "primary": "#b3842e",
    "accent": "#a1458e",
    "secondary": "#324a88",
    "success": "#3f802a",
    "info": "#79b1c9",
    "warning": "#b8721d",
    "error": "#a63228",
    "main": "#24221f",  # Are not traits
    "darker": "#1a1a1a",  # Are not traits
    "bg": "#121212",  # Are not traits
    "menu": "#424242",  # Are not traits
}
"colors used for the dark theme"

LIGHT_THEME: Dict[str, str] = {
    "primary": v.theme.themes.light.primary,
    "accent": v.theme.themes.light.accent,
    "secondary": v.theme.themes.light.secondary,
    "success": v.theme.themes.light.success,
    "info": v.theme.themes.light.info,
    "warning": v.theme.themes.light.warning,
    "error": v.theme.themes.light.error,
    "main": "#2e7d32",
    "darker": "#005005",
    "bg": "#FFFFFF",
    "menu": "#FFFFFF",
}
"colors used for the light theme"

TYPES: Tuple[str, ...] = (
    "info",
    "primary",
    "secondary",
    "accent",
    "error",
    "success",
    "warning",
    "anchor",
)
"The different types defined by ipyvuetify"

################################################################################
# define classes and method to make the application responsive
#


def get_theme() -> str:
    """
    Get theme name from the config file (default to dark).

    Returns:
        The theme to use
    """
    return config.get("sepal-ui", "theme", fallback="dark")


class SepalColor(HasTraits, SimpleNamespace):

    _dark_theme: Bool = Bool(True if get_theme() == "dark" else False).tag(sync=True)
    "Whether to use dark theme or not. By changing this value, the theme value will be stored in the conf file. Is only intended to be accessed in development mode."

    new_colors: dict = {}
    "Dictionary with name:color structure."

    @observe("_dark_theme")
    def __init__(self, *_, **new_colors) -> None:
        """
        Custom simple name space to store and access to the sepal_ui colors and with a magic method to display theme.

        Args:
            **new_colors (optional): the new colors to set in hexadecimal as a dict (experimetal)
        """
        # set vuetify theme
        v.theme.dark = self._dark_theme

        # Get get current theme name
        self.theme_name = "dark" if self._dark_theme else "light"

        # Save "new" theme in configuration file
        su.set_config("theme", self.theme_name)

        self.kwargs = DARK_THEME if self._dark_theme else LIGHT_THEME
        self.kwargs = new_colors or self.kwargs

        # Even if the theme.themes.dark_theme trait could trigger the change on all elms
        # we have to replace the default values everytime:
        theme = getattr(v.theme.themes, self.theme_name)

        # TODO: Would be awesome to find a way to create traits for the new colors and
        # assign them here directly
        [setattr(theme, color_name, color) for color_name, color in self.kwargs.items()]

        # Now instantiate the namespace
        SimpleNamespace.__init__(self, **self.kwargs)
        HasTraits.__init__(self)

        return

    def _repr_html_(self, *_) -> str:
        """Rich display of the color palette in an HTML frontend."""
        s = 60
        html = f"<h3>Current theme: {self.theme_name}</h3><table>"
        items = {k: v for k, v in self.kwargs.items()}.items()

        for name, color in items:
            c = su.to_colors(color)
            html += f"""
            <th>
                <svg width='{s}' height='{s}'>
                    <rect width='{s}' height='{s}' style='fill:{c};
                    stroke-width:1;stroke:rgb(255,255,255)'/>
                </svg>
            </th>
            """

        html += "</tr><tr>"
        html += "".join([f"<td>{name}</br>{color}</td>" for name, color in items])
        html += "</tr></table>"

        return html


# load custom styling of sepal_ui
sepal_ui_css = HTML(f"<style>{(CSS_DIR / 'custom.css').read_text()}</style>")

# load fa-6
fa_css = HTML(
    '<link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.2.1/css/all.min.css"/>'
)

# create a small hack to remove fontawesome from the html output
clean_fa_js = Javascript((JS_DIR / "fontawesome.js").read_text())

# display all
display(sepal_ui_css, fa_css, clean_fa_js)



"""
Module to gather all the models from the application. 

If you only have few widgets, a module is not necessary and you can simply use a model.py file
In a big module with lot of custom models, it can make sense to split things in separate files for the sake of maintenance.
If you use a module import all the functions here to only have 1 call to make
"""



"""
Package to customize the display of the widgets.

`IpyvuetfyWidget` and ìpyleaflet.Map` are not working together natively. This package provide helpers, css, js and js workaround to help user build safe application including both widgets and maps.   
"""



"""
Custom widgets that are nor input nor UI interface.

Gather the customized ``ipyvuetifyWidgets``. All the content of this modules is included in the parent ``sepal_ui.sepalwidgets`` package. So it can be imported directly from there.

Example:
    .. jupyter-execute::
    
        from sepal_ui import sepalwidgets as sw
        
        sw.CopyToClip()
"""

from pathlib import Path
from typing import Optional, Union

import ipyvuetify as v
import traitlets as t
from deprecated.sphinx import versionadded
from markdown import markdown
from traitlets import link, observe

from sepal_ui import color
from sepal_ui.model import Model
from sepal_ui.sepalwidgets.sepalwidget import SepalWidget, Tooltip

__all__ = ["Markdown", "CopyToClip", "StateIcon"]


class Markdown(v.Layout, SepalWidget):
    def __init__(self, mkd_str: str = "", **kwargs) -> None:
        """
        Custom Layout based on the markdown text given.

        Args:
            mkd_str: the text to display using the markdown convention. multi-line string are also interpreted
            kwargs: Any parameter from a v.Layout. If set, 'children' will be overwritten
        """
        mkd = markdown(mkd_str, extensions=["fenced_code", "sane_lists"])

        # need to be nested in a div to be displayed
        mkd = "<div>\n" + mkd + "\n</div>"

        # make every link to point to target black (to avoid nested iframe in sepal)
        mkd = mkd.replace("<a", '<a target="_blank"')

        # create a Html widget
        class MyHTML(v.VuetifyTemplate):
            template = t.Unicode(mkd).tag(sync=True)

        content = MyHTML()

        # set default parameters
        kwargs.setdefault("row", True)
        kwargs.setdefault("class_", "pa-5")
        kwargs.setdefault("align_center", True)
        kwargs["children"] = [v.Flex(xs12=True, children=[content])]

        # call the constructor
        super().__init__(**kwargs)


@versionadded(version="2.2.0", reason="New clipping widget")
class CopyToClip(v.VuetifyTemplate):

    tf: Optional[v.TextField] = None
    "v.TextField: the textfield widget that holds the v_model to copy"

    v_model: t.Unicode = t.Unicode("").tag(sync=True)
    "a v_model trait that embed the string to copy"

    def __init__(self, **kwargs) -> None:
        """
        Custom textField that provides a handy copy-to-clipboard javascript behaviour.

        When the clipboard btn is clicked the v_model will be copied in the local browser clipboard. You just have to change the clipboard v_model. when copied, the icon change from a copy to a check.

        Args:
            kwargs: any argument that can be used with a v.TextField
        """
        # add the default params to kwargs
        kwargs.setdefault("outlined", True)
        kwargs.setdefault("label", "Copy To clipboard")
        kwargs.setdefault("readonly", True)
        kwargs.setdefault("append_icon", "fa-solid fa-clipboard")
        kwargs.setdefault("v_model", "")
        kwargs.setdefault("class_", "ma-5")

        # set the default v_model
        self.v_model = kwargs["v_model"]

        # create component
        self.tf = v.TextField(**kwargs)
        self.components = {"mytf": self.tf}

        # template with js behaviour
        js_dir = Path(__file__).parents[1] / "frontend/js"
        clip = (js_dir / "jupyter_clip.js").read_text()
        self.template = (
            "<mytf/>" "<script>{methods: {jupyter_clip(_txt) {%s}}}</script>" % clip
        )

        super().__init__()

        # js behaviour
        self.tf.on_event("click:append", self._clip)
        link((self, "v_model"), (self.tf, "v_model"))

    def _clip(self, *args) -> None:
        """
        Launch the javascript clipping process.
        """
        self.send({"method": "clip", "args": [self.tf.v_model]})
        self.tf.append_icon = "fa-solid fa-clipboard-check"

        return


class StateIcon(Tooltip):

    values: t.Any = t.Any().tag(sync=True)
    "bool, str, int: key name of the current state of component. Values must be same as states_dict keys."

    states: dict = {}
    'Dictionary where keys are the state name to be linked with self value and value represented by a tuple of two elements. {"key":(tooltip_msg, color)}.'

    icon: Optional[v.Icon]
    "The colored Icon of the tooltip"

    def __init__(
        self,
        model: Union[None, Model] = None,
        model_trait: str = "",
        states: dict = {},
        **kwargs,
    ):
        """
        Custom icon with multiple state colors.

        Args:
            model: Model to manage StateIcon behaviour from outside.
            model_trait: Name of trait to be linked with state icon. Must exists in model.
            states: Dictionary where keys are the state name to be linked with self value and value represented by a tuple of two elements. {"key":(tooltip_msg, color)}.
            kwargs: Any arguments from a v.Tooltip
        """
        # set the default parameter of the tooltip
        kwargs.setdefault("right", True)

        # init the states
        default_states = {
            "valid": ("Valid", color.success),
            "non_valid": ("Not valid", color.error),
        }
        self.states = default_states if not states else states

        # Get the first value (states first key) to use as default one
        init_value = self.states[next(iter(self.states))]

        self.icon = v.Icon(
            children=["fa-solid fa-circle"], color=init_value[1], small=True
        )

        super().__init__(self.icon, init_value[0], **kwargs)

        # Directional from there to link here.
        if all([model, model_trait]):
            link((model, model_trait), (self, "values"))

    @observe("values")
    def _swap(self, change: dict) -> None:
        """Swap between states."""
        new_val = change["new"]

        # Use the first value when there is not initial value.
        if not new_val:
            self.value = next(iter(self.states))
            return

        # Perform a little check with comprehensive error message
        if new_val not in self.states:
            raise ValueError(
                f"Value '{new_val}' is not a valid value. Use {list(self.states.keys())}"
            )
        self.icon.color = self.states[new_val][1]
        self.children = [self.states[new_val][0]]

        return



"""
Customized control to display a legend on the map.
"""

from typing import Optional, Union

import traitlets as t
from ipyleaflet import WidgetControl
from ipywidgets import HTML
from traitlets import observe

import sepal_ui.sepalwidgets as sw
from sepal_ui.message import ms
from sepal_ui.scripts import utils as su


class LegendControl(WidgetControl):

    title: t.Unicode = t.Unicode(None).tag(sync=True)
    "Title of the legend."

    legend_dict: t.Dict = t.Dict(None).tag(sync=True)
    "Dictionary with key as label name and value as color"

    vertical: t.Bool = t.Bool(None).tag(sync=True)
    "Whether to display the legend in a vertical or horizontal way"

    _html_table: Optional[sw.Html] = None
    "The table containing the legend display"

    _html_title: Optional[sw.Html] = None
    "The tilte of the legend"

    def __init__(
        self,
        legend_dict: dict = {},
        title: str = ms.mapping.legend,
        vertical: bool = True,
        **kwargs,
    ) -> None:
        """
        A custom Legend widget ready to be embed in a map.

        This Legend can be control though it's different attributes, changin it's position of course but also the orientation ,the keys and their colors.

        .. versionadded:: 2.10.4

        Args:
            legend_dict: the dictionnary to fill the legend values. cannot be empty.
            title: title of the legend, if not set a default value in the current language will be used
            vertical: the orientation of the legend. default to True
        """
        # init traits
        self.title = title
        self.legend_dict = legend_dict
        self.vertical = vertical

        # generate the content based on the init options
        self._html_title = sw.Html(tag="h4", children=[f"{self.title}"])
        self._html_table = sw.Html(tag="table", children=[])

        # create a card inside the widget
        # Be sure that the scroll bar will be shown up when legend horizontal
        self.legend_card = sw.Card(
            attributes={"id": "legend_card"},
            style_="overflow-x:auto; white-space: nowrap;",
            max_width=450,
            max_height=350,
            children=[self._html_title, self._html_table],
        ).hide()

        # set some parameters for the actual widget
        kwargs["widget"] = self.legend_card
        kwargs.setdefault("position", "bottomright")

        super().__init__(**kwargs)

        self._set_legend(legend_dict)

    def __len__(self) -> int:
        """returns the number of elements in the legend."""
        return len(self.legend_dict)

    def hide(self) -> None:
        """Hide control by hiding its content."""
        self.legend_card.hide()

        return

    def show(self) -> None:
        """Show control by displaying its content."""
        self.legend_card.show()

        return

    @observe("legend_dict", "vertical")
    def _set_legend(self, *args) -> None:
        """Creates/update a legend based on the class legend_dict member."""
        # Do this to avoid crash when called by trait for the first time
        if self._html_table is None:
            return

        if not self.legend_dict:
            self.hide()
            return

        self.show()

        if self.vertical:
            elements = [
                sw.Html(
                    tag="tr" if self.vertical else "td",
                    children=[
                        sw.Html(tag="td", children=self.color_box(color)),
                        label.capitalize(),
                    ],
                )
                for label, color in self.legend_dict.items()
            ]
        else:
            elements = [
                (
                    sw.Html(
                        tag="td",
                        children=[label.capitalize()],
                    ),
                    sw.Html(
                        tag="td",
                        children=self.color_box(color),
                    ),
                )
                for label, color in self.legend_dict.items()
            ]
            # Flat nested list
            elements = [e for row in elements for e in row]

        self._html_table.children = elements

        return

    @observe("title")
    def _update_title(self, change: dict) -> None:
        """Trait method to update the title of the legend."""
        # Do this to avoid crash when called by trait
        if self._html_title is None:
            return

        self._html_title.children = change["new"]

        return

    @staticmethod
    def color_box(color: Union[str, tuple], size: int = 35) -> HTML:
        """
        Returns an rectangular SVG html element with the provided color.

        Args:
            color: It can be a string (e.g., 'red', '#ffff00', 'ffff00') or RGB tuple (e.g., (255, 127, 0))
            size: the size of the legend square

        Returns:
            The HTML rendered color box
        """
        # Define height and width based on the size
        w = size
        h = size / 2

        return [
            HTML(
                f"""
                    <th>
                        <svg width='{w}' height='{h}'>
                        <rect width='{w}' height='{h}' style='fill:{su.to_colors(color)};
                        stroke-width:1;stroke:rgb(255,255,255)'/>
                        </svg>
                    </th>
                """
            )
        ]



#!/usr/bin/python3

"""
Script to manually change the used theme.

This script will update the parameters shared between all sepal-ui based modules.
The "theme" parameter will be changed to the selected one. Any running application will need to be restarted to use this modification
"""

import argparse

from colorama import Fore, init

from sepal_ui.scripts import utils as su

# init colors for all plateforms
init()

# init parser
parser = argparse.ArgumentParser(description=__doc__, usage="module_theme")


def check_theme(theme: str) -> bool:
    """
    Check if the theme is a legit name.

    Args:
        theme: the theme name

    Returns:
        True if the theme is covered by sepal-ui
    """
    themes = ["dark", "light"]

    return theme in themes


def main() -> None:
    """Launch the process."""
    # parse agruments
    parser.parse_args()

    # welcome the user
    print(f"{Fore.YELLOW}sepal-ui module theme selection{Fore.RESET}")

    # select a language
    is_theme = False
    while is_theme is False:

        theme = input(f"{Fore.CYAN}Provide a theme name: \n{Fore.RESET}")
        is_theme = check_theme(theme)

        # display an error if the theme does not exist
        if is_theme is False:
            print(
                f'{Fore.RED} The provided theme name ("{theme}") is not a supported theme. {Fore.RESET}'
            )

    # write the new color code in the config file
    su.set_config_theme(theme)

    # display information
    print(
        f'{Fore.GREEN} The provided theme ("{theme}") has been set as default theme for all SEPAL applications.{Fore.RESET}'
    )

    return


if __name__ == "__main__":
    main()



"""
The customized ``Map`` object.
"""

# knwon bug of rasterio
import os

if "GDAL_DATA" in list(os.environ.keys()):
    del os.environ["GDAL_DATA"]
if "PROJ_LIB" in list(os.environ.keys()):
    del os.environ["PROJ_LIB"]

import json
import math
import random
import string
import warnings
from distutils.util import strtobool
from pathlib import Path
from typing import List, Optional, Sequence, Union, cast

import ee
import ipyleaflet as ipl
import ipyvuetify as v
import ipywidgets as widgets
import matplotlib.pyplot as plt
import numpy as np
import rioxarray
from deprecated.sphinx import deprecated
from matplotlib import colorbar
from matplotlib import colors as mpc
from rasterio.crs import CRS
from typing_extensions import Self

from sepal_ui import color as scolors
from sepal_ui import sepalwidgets as sw
from sepal_ui.frontend import styles as ss
from sepal_ui.mapping.basemaps import basemap_tiles
from sepal_ui.mapping.draw_control import DrawControl
from sepal_ui.mapping.layer import EELayer
from sepal_ui.mapping.layer_state_control import LayerStateControl
from sepal_ui.mapping.layers_control import LayersControl
from sepal_ui.mapping.legend_control import LegendControl
from sepal_ui.mapping.value_inspector import ValueInspector
from sepal_ui.mapping.zoom_control import ZoomControl
from sepal_ui.message import ms
from sepal_ui.scripts import decorator as sd
from sepal_ui.scripts import utils as su
from sepal_ui.scripts.warning import SepalWarning

__all__ = ["SepalMap"]


class SepalMap(ipl.Map):

    # ##########################################################################
    # ###                              Map parameters                        ###
    # ##########################################################################

    gee: bool = True
    "Either the map will use ee binding or not"

    v_inspector: Optional[ValueInspector] = None
    "The value inspector of the map"

    dc: Optional[DrawControl] = None
    "The drawing control of the map"

    _id: str = ""
    "A unique 6 letters str to identify the map in the DOM"

    state: Optional[sw.StateBar] = None
    "The statebar to inform the user about tile loading"

    def __init__(
        self,
        basemaps: List[str] = [],
        dc: bool = False,
        vinspector: bool = False,
        gee: bool = True,
        statebar: bool = False,
        **kwargs,
    ) -> None:
        """
        Custom Map object design to build application.

        The SepalMap class inherits from ipyleaflet.Map. It can thus be initialized with all its parameter.
        The map will fall back to CartoDB.DarkMatter map that well fits with the rest of the sepal_ui layout.
        Numerous methods have been added in the class to help you deal with your workflow implementation.
        It can natively display raster from .tif files and files and ee objects using methods
        that have the same signature as the GEE JavaScripts console.

        Args:
            basemaps: the basemaps used as background in the map. If multiple selection, they will be displayed as layers.
            dc: wether or not the drawing control should be displayed. default to false
            vinspector: Add value inspector to map, useful to inspect pixel values. default to false
            gee: wether or not to use the ee binding. If False none of the earthengine display fonctionalities can be used. default to True
            statebar: wether or not to display the Statebar in the map
            kwargs (optional): any parameter from a ipyleaflet.Map. if set, 'ee_initialize' will be overwritten.
        """
        # set the default parameters
        kwargs.setdefault("center", [0, 0])
        kwargs.setdefault("zoom", 2)
        kwargs["basemap"] = {}
        kwargs["zoom_control"] = False
        kwargs["attribution_control"] = False
        kwargs["scroll_wheel_zoom"] = True
        kwargs.setdefault("world_copy_jump", True)

        # Init the map
        super().__init__(**kwargs)

        # init ee
        self.gee = gee
        not gee or su.init_ee()

        # add the basemaps
        self.clear()
        default_basemap = (
            "CartoDB.DarkMatter" if v.theme.dark is True else "CartoDB.Positron"
        )
        basemaps = basemaps or [default_basemap]
        [self.add_basemap(basemap) for basemap in set(basemaps)]

        # set the visibility of all the basemaps to False but the first one
        [setattr(lyr, "visible", False) for lyr in self.layers]
        self.layers[0].visible = True

        # add the base controls
        self.add(ZoomControl(self))
        self.add(LayersControl(self, group=-1))
        self.add(ipl.AttributionControl(position="bottomleft", prefix="SEPAL"))
        self.add(ipl.ScaleControl(position="bottomleft", imperial=False))

        # specific drawing control
        self.dc = DrawControl(self)
        not dc or self.add(self.dc)

        # specific v_inspector
        self.v_inspector = ValueInspector(self)
        not vinspector or self.add(self.v_inspector)

        # specific statebar
        self.state = LayerStateControl(self)
        not statebar or self.add(self.state)

        # create a proxy ID to the element
        # this id should be unique and will be used by mutators to identify this map
        self._id = "".join(random.choice(string.ascii_lowercase) for i in range(6))
        self.add_class(self._id)

    @deprecated(version="2.8.0", reason="the local_layer stored list has been dropped")
    def _remove_local_raster(self, local_layer: str) -> Self:
        """
        Remove local layer from memory.

        .. danger::

            Does nothing now.

        Args:
            local_layer (str | ipyleaflet.TileLayer): The local layer to remove or its name
        """
        return self

    @deprecated(version="2.8.0", reason="use remove_layer(-1) instead")
    def remove_last_layer(self, local: bool = False) -> Self:
        """
        Remove last added layer from Map.

        Args:
            local: Specify True to only remove local last layers, otherwise will remove every last layer.
        """
        self.remove_layer(-1)

        return self

    def set_center(self, lon: float, lat: float, zoom: int = -1) -> None:
        """
        Centers the map view at a given coordinates with the given zoom level.

        Args:
            lon: The longitude of the center, in degrees.
            lat: The latitude of the center, in degrees.
            zoom: The zoom level, from 1 to 24. Defaults to None.
        """
        self.center = [lat, lon]
        self.zoom = self.zoom if zoom == -1 else zoom

        return

    @sd.need_ee
    def zoom_ee_object(self, item: ee.ComputedObject, zoom_out: int = 1) -> Self:
        """
        Get the proper zoom to the given ee geometry.

        Args:
            item: the geometry to zoom on
            zoom_out: Zoom out the bounding zoom
        """
        # type check the given object
        ee_geometry = item if isinstance(item, ee.Geometry) else item.geometry()

        # extract bounds from ee_object
        coords = ee_geometry.bounds().coordinates().get(0).getInfo()

        # zoom on these bounds
        return self.zoom_bounds((*coords[0], *coords[2]), zoom_out)

    def zoom_raster(self, layer: ipl.LocalTileLayer, zoom_out: int = 1) -> Self:
        """
        Adapt the zoom to the given LocalLayer.

        The localLayer need to come from the add_raster method to embed the image name.

        Args:
            layer: the localTile layer to zoom on. it needs to embed the "raster" member
            zoom_out: Zoom out the bounding zoom
        """
        da = rioxarray.open_rasterio(layer.raster, masked=True)

        # unproject if necessary
        epsg_4326 = "EPSG:4326"
        if da.rio.crs != CRS.from_string(epsg_4326):
            da = da.rio.reproject(epsg_4326)

        return self.zoom_bounds(da.rio.bounds(), zoom_out)

    def zoom_bounds(self, bounds: Sequence[float], zoom_out: int = 1) -> Self:
        """
        Adapt the zoom to the given bounds. and center the image.

        Args:
            bounds: coordinates corners as minx, miny, maxx, maxy in EPSG:4326
            zoom_out: Zoom out the bounding zoom
        """
        # center the map
        minx, miny, maxx, maxy = bounds
        self.fit_bounds([[miny, minx], [maxy, maxx]])

        # adapt the zoom level
        zoom_out = (self.zoom - 1) if zoom_out > self.zoom else zoom_out
        self.zoom -= zoom_out

        return self

    def add_raster(
        self,
        image: Union[str, Path],
        bands: Optional[Union[list, int]] = None,
        layer_name: str = "Layer_" + su.random_string(),
        colormap: Union[str, mpc.Colormap] = "inferno",
        opacity: float = 1.0,
        client_host: str = "/api/sandbox/jupyter/proxy/{port}",
        fit_bounds: bool = True,
    ) -> ipl.TileLayer:
        """
        Adds a local raster dataset to the map.

        Args:
            image: The image file path.
            bands: The image bands to use. It can be either a number (e.g., 1) or a list (e.g., [3, 2, 1]). Defaults to None.
            layer_name: The layer name to use for the raster. Defaults to None. If a layer is already using this name 3 random letter will be added
            colormap: The name of the colormap to use for the raster, such as 'gray' and 'terrain'. More can be found at https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html. Defaults to inferno.
            opacity: the opacity of the layer, default 1.0.
            client_host: the base url of the server. It's design to work in the SEPAL environment, you only need to change it if you want to work outside of our platform. See localtielayer lib for more details.
            fit_bounds: Wether or not we should fit the map to the image bounds. Default to True.

        Returns:
            the local tile layer embeding the raster member (to be used with other tools of sepal-ui)
        """
        # lazy import of localtileserver to avoid conflicts with GDAL
        # environments
        try:
            from localtileserver import TileClient, get_leaflet_tile_layer
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Your environment is not compatible with localtileserver, please check https://localtileserver.banesullivan.com/installation/index.html for more information"
            )

        # add the localtilelayer path to the environment
        # if the value is already set we don't change anything
        os.environ["LOCALTILESERVER_CLIENT_PREFIX"] = os.environ.pop(
            "LOCALTILESERVER_CLIENT_PREFIX", client_host
        )

        # force cast to Path and then start the client
        image = Path(image)

        if not image.is_file():
            raise Exception(ms.mapping.no_image)

        client = TileClient(image)

        # check inputs
        if layer_name in [layer.name for layer in self.layers]:
            layer_name = layer_name + su.random_string()

        # set the colors as independant colors
        if isinstance(colormap, str):
            cmap = plt.get_cmap(name=colormap)
        color_list = [mpc.rgb2hex(cmap(i)) for i in range(cmap.N)]

        da = rioxarray.open_rasterio(image, masked=True)
        da = da.chunk((1000, 1000))

        multi_band = False
        if len(da.band) > 1 and not isinstance(bands, int):
            multi_band = True
            bands = bands if bands else [3, 2, 1]
        elif len(da.band) == 1:
            bands = 1

        if multi_band:
            cast(list, bands)
            style = {
                "bands": [
                    {"band": bands[0], "palette": "#f00"},
                    {"band": bands[1], "palette": "#0f0"},
                    {"band": bands[2], "palette": "#00f"},
                ]
            }
        else:
            style = {
                "bands": [
                    {"band": bands, "palette": color_list},
                ]
            }

        # create the layer
        layer = get_leaflet_tile_layer(
            client, style=style, name=layer_name, opacity=opacity
        )
        self.add_layer(layer)

        # add the da to the layer as an extra member for the v_inspector
        layer.raster = str(image)

        # zoom on the layer if requested
        if fit_bounds is True:
            self.center = client.center()
            self.zoom = client.default_zoom

        return layer

    @deprecated(version="2.8.0", reason="use dc methods instead")
    def show_dc(self) -> Self:
        """Show the drawing control on the map."""
        self.dc.show()

        return self

    @deprecated(version="2.8.0", reason="use dc methods instead")
    def hide_dc(self) -> Self:
        """Hide the drawing control of the map."""
        self.dc.hide()

        return self

    def add_colorbar(
        self,
        colors: list,
        cmap: str = "viridis",
        vmin: float = 0.0,
        vmax: float = 1.0,
        index: list = [],
        categorical: bool = False,
        step: int = 0,
        transparent_bg: bool = False,
        position: str = "bottomright",
        layer_name: str = "",
        **kwargs,
    ) -> None:
        """
        Add a colorbar to the map.

        Args:
            colors: The set of colors to be used for interpolation. Colors can be provided in the form: * tuples of RGBA ints between 0 and 255 (e.g: (255, 255, 0) or (255, 255, 0, 255)) * tuples of RGBA floats between 0. and 1. (e.g: (1.,1.,0.) or (1., 1., 0., 1.)) * HTML-like string (e.g: “#ffff00) * a color name or shortcut (e.g: “y” or “yellow”)
            cmap: a matplotlib colormap default to viridis
            vmin: The minimal value for the colormap. Values lower than vmin will be bound directly to colors[0].. Defaults to 0.
            vmax: The maximal value for the colormap. Values higher than vmax will be bound directly to colors[-1]. Defaults to 1.0.
            index: The values corresponding to each color. It has to be sorted, and have the same length as colors. If None, a regular grid between vmin and vmax is created. Defaults to None.
            categorical (bool, optional): Whether or not to create a categorical colormap. Defaults to False.
            step: The step to split the LinearColormap into a StepColormap. Defaults to None.
            position: The position for the colormap widget. Defaults to "bottomright".
            layer_name: Layer name of the colorbar to be associated with. Defaults to None.
            kwargs: any other argument of the colorbar object from matplotlib
        """
        width, height = 6.0, 0.4
        alpha = 1

        if colors is not None:

            # transform colors in hex colors
            hexcodes = [su.to_colors(c) for c in colors]

            if categorical:
                plot_color = mpc.ListedColormap(hexcodes)
                vals = np.linspace(vmin, vmax, plot_color.N + 1)
                norm = mpc.BoundaryNorm(vals, plot_color.N)

            else:
                plot_color = mpc.LinearSegmentedColormap.from_list(
                    "custom", hexcodes, N=256
                )
                norm = mpc.Normalize(vmin=vmin, vmax=vmax)

        elif cmap is not None:

            plot_color = plt.get_cmap(cmap)
            norm = mpc.Normalize(vmin=vmin, vmax=vmax)

        else:
            msg = '"cmap" keyword or "colors" key must be provided.'
            raise ValueError(msg)

        style = "dark_background" if v.theme.dark is True else "classic"

        with plt.style.context(style):
            fig, ax = plt.subplots(figsize=(width, height))
            cb = colorbar.ColorbarBase(
                ax,
                norm=norm,
                alpha=alpha,
                cmap=plot_color,
                orientation="horizontal",
                **kwargs,
            )

            # cosmetic improvement
            cb.outline.set_visible(False)  # remove border of the color bar
            ax.tick_params(size=0)  # remove ticks
            fig.patch.set_alpha(0.0)  # remove bg of the fig
            ax.patch.set_alpha(0.0)  # remove bg of the ax

        not layer_name or cb.set_label(layer_name)

        output = widgets.Output()
        colormap_ctrl = ipl.WidgetControl(
            widget=output,
            position=position,
            transparent_bg=True,
        )
        with output:
            output.clear_output()
            plt.show()

        self.add(colormap_ctrl)

        return

    def add_ee_layer(
        self,
        ee_object: ee.ComputedObject,
        vis_params: dict = {},
        name: str = "",
        shown: bool = True,
        opacity: float = 1.0,
        viz_name: str = "",
    ) -> None:
        """
        Customized add_layer method designed for EE objects.

        Copy the addLayer method from geemap to read and guess the vizaulization
        parameters the same way as in SEPAL recipes.
        If the vizparams are empty and vizualization metadata exist, SepalMap will use
        them automatically.

        Args:
            ee_object: the ee OBject to draw on the map
            vis_params: the visualization parameters set as in GEE
            name: the name of the layer
            shown: either to show the layer or not, default to true (it is bugged in ipyleaflet)
            opacity: the opcity of the layer from 0 to 1, default to 1.
            viz_name: the name of the vizaulization you want ot use. default to the first one if existing
        """
        # check the type of the ee object and raise an error if it's not recognized
        if not isinstance(
            ee_object,
            (
                ee.Image,
                ee.ImageCollection,
                ee.FeatureCollection,
                ee.Feature,
                ee.Geometry,
            ),
        ):
            raise AttributeError(
                "\n\nThe image argument in 'addLayer' function must be an instance of "
                "one of ee.Image, ee.Geometry, ee.Feature or ee.FeatureCollection."
            )

        # get the list of viz params
        viz = self.get_viz_params(ee_object)

        # get the requested vizparameters name
        # if non is set use the first one
        if viz:
            viz_name = viz_name or viz[next(iter(viz))]["name"]

        # apply it to vis_params
        if not vis_params and viz:

            # find the viz params in the list
            try:
                vis_params = next(i for p, i in viz.items() if i["name"] == viz_name)
            except StopIteration:
                raise ValueError(
                    f"the provided viz_name ({viz_name}) cannot be found in the image metadata"
                )

            # invert the bands if needed
            inverted = vis_params.pop("inverted", None)
            if inverted is not None:

                # get the index of the bands that need to be inverted
                index_list = [i for i, v in enumerate(inverted) if v is True]

                # multiply everything by -1
                for i in index_list:
                    min_ = vis_params["min"][i]
                    max_ = vis_params["max"][i]
                    vis_params["min"][i] = max_
                    vis_params["max"][i] = min_

            # specific case of categorical images
            # Pad the palette when using non-consecutive values
            # instead of remapping or using sldStyle
            # to preserve the class values in the image, for inspection
            if vis_params["type"] == "categorical":

                colors = vis_params["palette"]
                values = vis_params["values"]
                min_ = min(values)
                max_ = max(values)

                # set up a black palette of correct length
                palette = ["#000000"] * (max_ - min_ + 1)

                # replace the values within the palette
                for i, val in enumerate(values):
                    palette[val - min_] = colors[i]

                # adapt the vizparams
                vis_params["palette"] = palette
                vis_params["min"] = min_
                vis_params["max"] = max_

            # specific case of hsv
            elif vis_params["type"] == "hsv":

                # set to_min to 0 and to_max to 1
                # in the original expression:
                # 'to_min + (v - from_min) * (to_max - to_min) / (from_max - from_min)'
                expression = (
                    "{band} = (b('{band}') - {from_min}) / ({from_max} - {from_min})"
                )

                # get the maxs and mins
                # removing them from the parameter
                mins = vis_params.pop("min")
                maxs = vis_params.pop("max")

                # create the rgb bands
                asset = ee_object
                for i, band in enumerate(vis_params["bands"]):

                    # adapt the expression
                    exp = expression.format(
                        from_min=mins[i], from_max=maxs[i], band=band
                    )
                    asset = asset.addBands(asset.expression(exp), [band], True)

                # set the arguments
                ee_object = asset.select(vis_params["bands"]).hsvToRgb()
                vis_params["bands"] = ["red", "green", "blue"]

        # create the layer based on these new values
        if not name:
            layer_count = len(self.layers)
            name = "Layer " + str(layer_count + 1)

        # force cast to featureCollection if needed
        if isinstance(
            ee_object,
            (
                ee.geometry.Geometry,
                ee.feature.Feature,
                ee.featurecollection.FeatureCollection,
            ),
        ):
            default_vis = json.loads((ss.JSON_DIR / "layer.json").read_text())[
                "ee_layer"
            ]
            default_vis.update(color=scolors.primary)

            # We want to get all the default styles and only change those whose are
            # in the provided visualization.
            default_vis.update(vis_params)

            vis_params = default_vis

            features = ee.FeatureCollection(ee_object)
            const_image = ee.Image.constant(0.5)

            try:
                image_fill = features.style(**vis_params).updateMask(const_image)
                image_outline = features.style(**vis_params)

            except AttributeError:
                # Raise a more understandable error
                raise AttributeError(
                    "You can only use the following styles: 'color', 'pointSize', "
                    "'pointShape', 'width', 'fillColor', 'styleProperty', "
                    "'neighborhood', 'lineType'"
                )

            image = image_fill.blend(image_outline)
            obj = features

        # use directly the ee object if Image
        elif isinstance(ee_object, ee.image.Image):
            image = obj = ee_object

        # use mosaicing if the ee_object is a ImageCollection
        elif isinstance(ee_object, ee.imagecollection.ImageCollection):
            image = obj = ee_object.mosaic()

        # create the colored image
        map_id_dict = ee.Image(image).getMapId(vis_params)
        tile_layer = EELayer(
            ee_object=obj,
            url=map_id_dict["tile_fetcher"].url_format,
            attribution="Google Earth Engine",
            name=name,
            opacity=opacity,
            visible=shown,
            max_zoom=24,
        )

        self.add_layer(tile_layer)

        return

    @staticmethod
    def get_basemap_list() -> List[str]:
        """
        Get the complete list of avaialble basemaps.

        This function is intending for development use
        It give the list of all the available basemaps for SepalMap object.

        Returns:
            The list of the basemap names
        """
        return [k for k in basemap_tiles.keys()]

    @staticmethod
    def get_viz_params(image: ee.Image) -> dict:
        """
        Return the vizual parameters that are set in the metadata of the image.

        Args:
            image: the image to analyse

        Returns:
            The dictionnary of the find properties
        """
        # the constant prefix for SEPAL visualization parameters
        PREFIX = "visualization"

        # init the property list
        props = {}

        # check image type
        if not isinstance(image, ee.Image):
            return props

        # check that image have properties
        if "properties" not in image.getInfo():
            return props

        # build a raw prop list
        raw_prop_list = {
            p: val
            for p, val in image.getInfo()["properties"].items()
            if p.startswith(PREFIX)
        }

        # decompose each property by its number
        # and gather the properties in a sub dictionnary
        for p, val in raw_prop_list.items():

            # extract the number and create the sub-dict
            _, number, name = p.split("_")
            props.setdefault(number, {})

            # modify the values according to prop key
            if isinstance(val, str):
                if name in ["bands", "palette", "labels"]:
                    val = val.split(",")
                elif name in ["max", "min", "values"]:
                    val = [float(i) for i in val.split(",")]
                elif name in ["inverted"]:
                    val = [bool(strtobool(i)) for i in val.split(",")]

            # set the value
            props[number][name] = val

        for i in props.keys():
            if "type" in props[i]:
                # categorical values need to be cast to int
                if props[i]["type"] == "categorical":
                    props[i]["values"] = [int(val) for val in props[i]["values"]]
            else:
                # if no "type" is provided guess it from the different parameters gathered
                if len(props[i]["bands"]) == 1:
                    props[i]["type"] = "continuous"
                elif len(props[i]["bands"]) == 3:
                    props[i]["type"] = "rgb"
                else:
                    warnings.warn(
                        "the embed viz properties are incomplete or badly set, "
                        "please review our documentation",
                        SepalWarning,
                    )
                    props = {}

        return props

    def remove_layer(
        self, key: Union[ipl.Layer, int, str], base: bool = False, none_ok: bool = False
    ) -> None:
        """
        Remove a layer based on a key.

        The key can be, a Layer object, the name of a layer or the index in the layer list.

        Args:
            key: the key to find the layer to delete
            base: either the basemaps should be included in the search or not. default t false
            none_ok: if True the function will not raise error if no layer is found. Default to False
        """
        layer = self.find_layer(key, base, none_ok)

        # the error is catched in find_layer
        if layer is not None:
            super().remove(layer)

        return

    def remove_all(self, base: bool = False) -> None:
        """
        Remove all the layers from the maps.

        If base is set to True, the basemaps are removed as well.

        Args:
            base: wether or not the basemaps should be removed, default to False
        """
        # filter out the basemaps if base == False
        layers = self.layers if base else [lyr for lyr in self.layers if not lyr.base]

        # remove them using the layer objects as keys
        [self.remove_layer(layer, base) for layer in layers]

        return

    def add_layer(self, layer: ipl.Layer, hover: bool = False) -> None:
        """
        Add layer and use a default style for the GeoJSON inputs.

        Remove existing layer if already on the map.

        Args:
            layer: any layer type from ipyleaflet
            hover: whether to use the default hover style or not.
        """
        # remove existing layer before addition
        existing_layer = self.find_layer(layer.name, none_ok=True)
        not existing_layer or self.remove_layer(existing_layer)

        # apply default coloring for geoJson
        if isinstance(layer, ipl.GeoJSON):

            # define the default values
            default_style = json.loads((ss.JSON_DIR / "layer.json").read_text())[
                "layer"
            ]
            default_style.update(color=scolors.primary)
            default_hover_style = json.loads(
                (ss.JSON_DIR / "layer_hover.json").read_text()
            )
            default_hover_style.update(color=scolors.primary)

            # apply the style depending on the parameters
            layer.style = layer.style or default_style
            hover_style = default_hover_style if hover else layer.hover_style
            layer.hover_style = layer.hover_style or hover_style

        super().add(layer)

        return

    def add_basemap(self, basemap: str = "HYBRID") -> None:
        """
        Adds a basemap to the map.

        Args:
            basemap: Can be one of string from basemaps. Defaults to 'HYBRID'.
        """
        if basemap not in basemap_tiles.keys():
            keys = "\n".join(basemap_tiles.keys())
            msg = f"Basemap can only be one of the following:\n{keys}"
            raise ValueError(msg)

        self.add_layer(basemap_tiles[basemap])

        return

    def get_scale(self) -> float:
        """
        Returns the approximate pixel scale of the current map view, in meters.

        Reference: https://blogs.bing.com/maps/2006/02/25/map-control-zoom-levels-gt-resolution.

        Returns:
            Map resolution in meters.
        """
        return 156543.04 * math.cos(0) / math.pow(2, self.zoom)

    def find_layer(
        self, key: Union[ipl.Layer, str, int], base: bool = False, none_ok: bool = False
    ) -> ipl.TileLayer:
        """
        Search a layer by name or index.

        Args:
            key: the layer name, index or directly the layer
            base: either the basemaps should be included in the search or not. default to false
            none_ok: if True the function will not raise error if no layer is found. Default to False

        Returns:
            The first layer using the same name or index else None
        """
        # filter the layers
        layers = self.layers if base else [lyr for lyr in self.layers if not lyr.base]

        if isinstance(key, str):
            layer = next((lyr for lyr in layers if lyr.name == key), None)
        elif isinstance(key, int):
            size = len(layers)
            layer = layers[key] if -size <= key < size else None
        elif isinstance(key, ipl.Layer):
            layer = next((lyr for lyr in layers if lyr == key), None)
        else:
            raise ValueError(f"key must be a int or a str, {type(key)} given")

        if layer is None and none_ok is False:
            raise ValueError(f"no layer corresponding to {key} on the map")

        return layer

    def add_legend(
        self,
        title: str = ms.mapping.legend,
        legend_dict: dict = {},
        position: str = "bottomright",
        vertical: bool = True,
    ) -> None:
        """
        Creates and adds a custom legend as widget control to the map.

        Args:
            title: Title of the legend. Defaults to 'Legend'.
            legend_dict: dictionary with key as label name and value as color
            position: the position (corners) of the legend on the map
            vertical: vertical or horizoal position of the legend
        """
        # Define as class member so it can be accessed from outside.
        self.legend = LegendControl(
            legend_dict, title=title, vertical=vertical, position=position
        )

        return self.add(self.legend)

    # ##########################################################################
    # ###                overwrite geemap calls                              ###
    # ##########################################################################

    setCenter = set_center
    centerObject = zoom_ee_object
    addLayer = add_ee_layer
    getScale = get_scale



"""
Gather all the scripts used to support the sepal-ui interface (widgets and map).

Example:
    .. jupyter-execute::
    
        from sepal_ui.scripts import utils as su
        
        su.random_string(5)
"""



"""
wrapper for ipyvuetify widgets to unify the display of voila dashboards in the SEPAL plateform.

``sepal_ui`` is a lib designed to create elegant python based dashboard in the SEPAL environment. It is designed on top of the amazing ``ipyvuetify`` library and will help developer to easily create interface for their workflows. By using this libraries, you'll ensure a robust and unified interface for your scripts and a easy and complete integration into the SEPAL dashboard of application.
"""

from sepal_ui.conf import config as config
from sepal_ui.conf import config_file as config_file
from sepal_ui.frontend.styles import SepalColor
from sepal_ui.frontend.styles import get_theme as get_theme

__author__ = """Pierrick Rambaud"""
__email__ = "pierrick.rambaud49@gmail.com"
__version__ = "2.15.0"

color = SepalColor()
'color: the colors of sepal. members are in the following list: "main, darker, bg, primary, accent, secondary, success, info, warning, error, menu". They will render according to the selected theme.'



"""
module to gather all the parameters used in the application. 

if you only have few widgets, a module is not necessary and you can simply use a parameter.py file. Iin a big module with lot of custom parmeters, it can make sense to split things in separate files for the sake of maintenance.
If you use a module import all the functions here to only have 1 call to make
"""

from .directory import *



"""
Custom widgets relative to user outputs.

Gather the customized ``ipyvuetifyWidgets`` used to communicate with the end user.
All the content of this modules is included in the parent ``sepal_ui.sepalwidgets`` package. So it can be imported directly from there.

Example:
    .. jupyter-execute::
    
        from sepal_ui import sepalwidgets as sw
        
        sw.Alert().show()
"""

from datetime import datetime
from typing import Any, Optional

import ipyvuetify as v
import traitlets as t
from deprecated.sphinx import deprecated
from ipywidgets import Output, jslink
from tqdm.notebook import tqdm
from traitlets import directional_link, observe
from typing_extensions import Self

from sepal_ui import color
from sepal_ui.frontend.styles import TYPES
from sepal_ui.message import ms
from sepal_ui.scripts import utils as su
from sepal_ui.scripts.utils import set_type
from sepal_ui.sepalwidgets.sepalwidget import SepalWidget

__all__ = ["Divider", "Alert", "StateBar", "Banner"]


class Divider(v.Divider, SepalWidget):

    type_: t.Unicode = t.Unicode("").tag(sync=True)
    "Added type\_ trait to specify the current color of the divider"

    def __init__(self, class_: str = "", **kwargs) -> None:
        r"""
        A custom Divider with the ability to dynamically change color.

        Whenever the type\_ trait is modified, the divider class will change accordingly.

        Args:
            class\_: the initial color of the divider
            kwargs (optional): any parameter from a v.Divider. if set, 'class\_' will be overwritten.
        """
        kwargs["class_"] = class_
        super().__init__(**kwargs)

    @observe("type_")
    def add_class_type(self, change: dict) -> Self:
        r"""
        Change the color of the divider according to the type\_.

        It is binded to the type\_ traitlet but can also be called manually.

        Args:
            change: the only useful key is 'new' which is the new required color
        """
        self.class_list.remove(*TYPES)
        self.class_list.add(change["new"])

        return self


class Alert(v.Alert, SepalWidget):

    progress_bar: Optional[tqdm] = None
    "the progress bar of the alert"

    progress_output: Optional[Output] = None
    "the output object where the progress bar is stored"

    def __init__(self, type_: str = "info", **kwargs) -> None:
        r"""
        A custom Alert widget.

        It is used as the output of all processes in the framework.
        In the voila interfaces, print statement will not be displayed.
        Instead use the sw.Alert method to provide information to the user.
        It's hidden by default.

        Args:
            type\_: The color of the Alert
            kwargs (optional): any parameter from a v.Alert. If set, 'type' will be overwritten.
        """
        # set default parameters
        kwargs.setdefault("text", True)
        kwargs["type"] = set_type(type_)
        kwargs.setdefault("class_", "mt-5")

        # call the constructor
        super().__init__(**kwargs)

        self.hide()
        self.progress_output = Output()
        # self.progress_bar = None

    def update_progress(
        self, progress: float, msg: str = "Progress", **tqdm_args
    ) -> None:
        """
        Update the Alert message with a tqdm progress bar.

        .. note::

            set the ``total`` argumentent of tqdm to use differnet values than [0, 1]

        Args:
            progress: the progress status in float
            msg: The message to use before the progress bar
            tqdm_args (optional): any arguments supported by a tqdm progress bar
        """
        # show the alert
        self.show()

        # cast the progress to float
        total = tqdm_args.get("total", 1)
        progress = float(progress)
        if not (0 <= progress <= total):
            raise ValueError(f"progress should be in [0, {total}], {progress} given")

        # Prevent adding multiple times
        if self.progress_output not in self.children:

            self.children = [self.progress_output]

            tqdm_args.setdefault("bar_format", "{l_bar}{bar}{n_fmt}/{total_fmt}")
            tqdm_args.setdefault("dynamic_ncols", False)
            tqdm_args.setdefault("total", 1)
            tqdm_args.setdefault("desc", msg)
            tqdm_args.setdefault("colour", getattr(color, self.type))

            with self.progress_output:
                self.progress_output.clear_output()
                self.progress_bar = tqdm(**tqdm_args)
                self.progress_bar.container.children[0].add_class(f"{self.type}--text")
                self.progress_bar.container.children[2].add_class(f"{self.type}--text")

                # Initialize bar
                self.progress_bar.update(0)

        self.progress_bar.update(progress - self.progress_bar.n)

        if progress == 1:
            self.progress_bar.close()

        return

    def add_msg(self, msg: str, type_: str = "info") -> Self:
        r"""
        Add a message in the alert by replacing all the existing one.

        The color can also be changed dynamically.

        Args:
            msg: the message to display
            type\_: the color to use in the widget
        """
        self.show()
        self.type = set_type(type_)
        self.children = [v.Html(tag="p", children=[msg])]

        return self

    def add_live_msg(self, msg: str, type_: str = "info") -> Self:
        r"""
        Add a message in the alert by replacing all the existing one.

        Also add the timestamp of the display. The color can also be changed dynamically.

        Args:
            msg: the message to display
            type\_: the color to use in the widget
        """
        current_time = datetime.now().strftime("%Y/%m/%d, %H:%M:%S")

        self.show()
        self.type = set_type(type_)
        self.children = [
            v.Html(tag="p", children=["[{}]".format(current_time)]),
            v.Html(tag="p", children=[msg]),
        ]

        return self

    def append_msg(self, msg: str, section: bool = False, type_: str = "info") -> Self:
        r"""
        Append a message in a new parragraph, with or without divider.

        Args:
            msg: the message to display
            section: add a Divider before the added message
            type\_: the color to use in the widget
        """
        self.show()
        self.type = type_

        if len(self.children):
            current_children = self.children[:]
            if section:
                # As the list is mutable, and the trait is only triggered when
                # the children is changed, so have to create a copy.
                divider = Divider(class_="my-4", style="opacity: 0.22")

                # link Alert type with divider type
                directional_link((self, "type"), (divider, "type_"))
                current_children.extend([divider, v.Html(tag="p", children=[msg])])
                self.children = current_children

            else:
                current_children.append(v.Html(tag="p", children=[msg]))
                self.children = current_children
        else:
            self.add_msg(msg)

        return self

    def remove_last_msg(self) -> Self:
        """
        Remove the last msg printed in the Alert widget.
        """
        if len(self.children) > 1:
            current_children = self.children[:]
            self.children = current_children[:-1]
        else:
            self.reset()

        return self

    def reset(self) -> Self:
        """
        Empty the messages and hide it.
        """
        self.children = [""]
        self.hide()

        return self

    @deprecated(version="3.0", reason="This method is now part of the utils module")
    def check_input(self, input_: Any, msg: str = "") -> bool:
        r"""
        Check if the inpupt value is initialized.

        If not return false and display an error message else return True.

        Args:
            input\_: the input to check
            msg: the message to display if the input is not set

        Returns:
            check if the value is initialized
        """
        msg = msg or ms.utils.check_input.error

        return su.check_input(input_, msg)


class StateBar(v.SystemBar, SepalWidget):

    msg: t.Unicode = t.Unicode("").tag(sync=True)
    "the displayed message"

    loading: t.Bool = t.Bool(False).tag(sync=True)
    "either or not the circular progress is spinning"

    progress: Optional[v.ProgressCircular] = None
    "The ProgressCircular widget that will be displayed in the statebar"

    def __init__(self, **kwargs) -> None:
        """
        Widget to display quick messages on simple inline status bar.

        Args:
            kwargs (optional): any parameter from a v.SystemBar. If set, 'children' will be overwritten.
        """
        self.progress = v.ProgressCircular(
            indeterminate=self.loading,
            value=100,
            small=True,
            size=15,
            color="primary",
            class_="mr-2",
        )

        # set default parameter
        kwargs["children"] = [self.progress, self.msg]

        # call the constructor
        super().__init__(**kwargs)

        jslink((self, "loading"), (self.progress, "indeterminate"))

    @observe("loading")
    def _change_loading(self, *args) -> None:
        """Change progress wheel state."""
        self.progress.indeterminate = self.loading

        return

    @observe("msg")
    def _change_msg(self, *args) -> None:
        """Change state bar message."""
        self.children = [self.progress, self.msg]

        return

    def add_msg(self, msg: str, loading: bool = False) -> Self:
        """
        Change current status message.

        Args:
            msg: the message to display
            loading: the loading status of the progress circular
        """
        self.msg = msg
        self.loading = loading

        return self


class Banner(v.Snackbar, SepalWidget):

    btn_close: Optional[v.Btn] = None
    "v.Btn: the closing btn of the banner"

    def __init__(
        self,
        msg: str = "",
        type_: str = "info",
        id_: str = "",
        persistent: bool = True,
        **kwargs,
    ) -> None:
        r"""
        Custom Snackbar widget to display messages as a banner in module App.

        Args:
            msg: Message to display in application banner. default to nothing
            type\_: Used to display an appropiate banner color. fallback to "info".
            id_: unique banner identificator.
            persistent: Whether to close automatically based on the lenght of message (False) or make it indefinitely open (True). Overridden if timeout duration is set.
            kwargs (optional): any parameter from a v.Alert. If set, 'vertical' and 'top' will be overwritten.
        """
        # compute the type and default to "info" if it's not existing
        type_ = set_type(type_)

        # create the closing btn
        self.btn_close = v.Btn(
            small=True, text=True, children=[ms.widgets.banner.close]
        )

        # compute timeout based on the persistent and timeout parameter
        computed_timeout = 0 if persistent is True else self.get_timeout(msg)

        kwargs.setdefault("color", type_)
        kwargs.setdefault("transition", "scroll-x-transition")
        kwargs["attributes"] = {"id": id_}
        kwargs.setdefault("v_model", True)
        kwargs.setdefault("timeout", computed_timeout)
        kwargs["top"] = True
        kwargs["vertical"] = True
        kwargs["children"] = [msg] + [self.btn_close]
        kwargs["class_"] = "mb-1"

        super().__init__(**kwargs)

        self.btn_close.on_event("click", self.close)

    def close(self, *args) -> None:
        """Close button event to close snackbar alert."""
        self.v_model = False  # mypy: ignore-errors

        return

    def get_timeout(self, text: str) -> int:
        """
        Calculate timeout in miliseconds to read the message.

        Args:
            text: the text displayed in the banner to adapt the duration of the timeout

        Returns:
            the duration of the timeout in milliseconds
        """
        wpm = 180  # readable words per minute
        word_length = 5  # standardized number of chars in calculable word
        words = len(text) / word_length
        words_time = ((words / wpm) * 60) * 1000

        delay = 1500  # milliseconds before user starts reading the notification
        bonus = 1000  # extra time

        return int(delay + words_time + bonus)

    def set_btn(self, nb_banner: int) -> None:
        """
        Change the btn display to inform the user on the number of banners in the queue.

        Args:
            nb_banner: the number of banners in the queue
        """
        # do not wrap ms.widget.banner. If you do it won't be recognized by the key-checker of the Translator
        if nb_banner == 0:
            txt = ms.widgets.banner.close
        else:
            txt = ms.widgets.banner.next.format(nb_banner)
        self.btn_close.children = [txt]

        return



"""
Extend fonctionalities of the ipyleaflet layer control.
"""
import json
from types import SimpleNamespace
from typing import Optional

import ipyvuetify as v
from ipyleaflet import GeoJSON, Map, TileLayer
from ipywidgets import link

from sepal_ui import color
from sepal_ui import sepalwidgets as sw
from sepal_ui.frontend import styles as ss
from sepal_ui.mapping.menu_control import MenuControl
from sepal_ui.message import ms


class HeaderRow(sw.Html):
    def __init__(self, title: str) -> None:
        """
        Html Row element including a single title of 3 colspan.

        Specifically designed to work in the layer_control table

        Args:
            title: the line value
        """
        attr = {"colspan": 3}
        head = sw.Html(tag="th", attributes=attr, children=title)
        super().__init__(tag="tr", class_="v-no-hover", children=[head])


class BaseRow(sw.Html):

    w_radio: Optional[sw.SimpleCheckbox] = None
    "the radio to hide/show the layer"

    def __init__(self, layer: TileLayer) -> None:
        """
        Html row element to describe a base layer.

        This Html element include all the controls to manipulate the basemap displayed in the map
        - a checkbox to show/hide (it will behave like a radio)

        Args:
            layer: the layer associated to the row
        """
        # create the checkbox, by default layer are visible
        self.w_radio = sw.Radio(small=True, value=layer.name)
        kwargs = {"style": "width: 10%;", "tag": "td"}
        radio_cell = sw.Html(children=[self.w_radio], **kwargs)

        # create the label
        kwargs = {"style_": "width: 40%;", "tag": "td"}
        label_cell = sw.Html(children=[layer.name], **kwargs)

        # create an empty row to align on
        kwargs = {"style_": "width: 50%;", "tag": "td"}
        empty_cell = sw.Html(children=[""], **kwargs)

        # build a html tr from it
        super().__init__(tag="tr", children=[label_cell, empty_cell, radio_cell])

        # add js behavior
        link((layer, "visible"), (self.w_radio, "active"))


class LayerRow(sw.Html):

    w_checkbox: Optional[sw.SimpleCheckbox] = None
    "the ckeckbox to hide/show the layer"

    w_slider: Optional[sw.SimpleSlider] = None
    "the slider linked to the opacity of the layer"

    def __init__(self, layer: TileLayer) -> None:
        """
        Html row element to describe a normal layer.

        This Html element include all the controls to manipulate the layer displayed in the map
        - a checkbox to show/hide
        - a slider to change alpha channel

        Args:
            layer: the layer associated to the row
        """
        # create the checkbox, by default layer are visible
        self.w_checkbox = sw.SimpleCheckbox(
            v_model=True, small=True, label=layer.name, color=color.primary
        )
        kwargs = {"style": "width: 10%;", "tag": "td"}
        checkbox_cell = sw.Html(children=[self.w_checkbox], **kwargs)

        # create the label
        kwargs = {"style_": "width: 40%;", "tag": "td"}
        label_cell = sw.Html(children=[layer.name], **kwargs)

        # create the slider
        self.w_slider = sw.SimpleSlider(v_model=1, max=1, step=0.01, small=True)
        kwargs = {"style_": "width: 50%;", "tag": "td"}
        slider_cell = sw.Html(children=[self.w_slider], **kwargs)

        # build a html tr from it
        super().__init__(tag="tr", children=[label_cell, slider_cell, checkbox_cell])

        # add js behavior
        link((layer, "opacity"), (self.w_slider, "v_model"))
        link((self.w_checkbox, "v_model"), (layer, "visible"))
        self.w_checkbox.observe(self._toggle_slider, "v_model")

    def _toggle_slider(self, *args) -> None:
        """Toggle the modification of the slider."""
        self.w_slider.disabled = not self.w_checkbox.v_model

        return


class VectorRow(sw.Html):

    w_checkbox: Optional[sw.SimpleCheckbox] = None
    "the ckeckbox to hide/show the layer"

    def __init__(self, layer: TileLayer) -> None:
        """
        Html row element to describe a vector layer.

        This Html element include all the controls to manipulate the layer displayed in the map
        - a checkbox to show/hide
        Vector are always placed on top of the map

        Args:
            layer: the vector layer associated to the row
        """
        # create the checkbox, by default layer are visible
        self.w_checkbox = sw.SimpleCheckbox(
            v_model=True, small=True, label=layer.name, color=color.primary
        )
        kwargs = {"style": "width: 10%;", "tag": "td"}
        checkbox_cell = sw.Html(children=[self.w_checkbox], **kwargs)

        # create the label
        kwargs = {"style_": "width: 40%;", "tag": "td"}
        label_cell = sw.Html(children=[layer.name], **kwargs)

        # create the slider
        kwargs = {"style_": "width: 50%;", "tag": "td"}
        empty_cell = sw.Html(children=[""], **kwargs)

        # build a html tr from it
        super().__init__(tag="tr", children=[label_cell, empty_cell, checkbox_cell])

        # add js behavior
        link((self.w_checkbox, "v_model"), (layer, "visible"))


class LayersControl(MenuControl):

    m: Optional[Map] = None
    "the map controlled by the layercontrol"

    group: Optional[sw.RadioGroup] = None
    "As radio button cannot behave individually we add an extra GroupRadio to wrap the table"

    def __init__(self, m: Map, **kwargs) -> None:
        """
        Richer layerControl to add some controls over the lyers displayed on the map.

        Each layer is associated to a line where the user can adapt the alpha chanel or even hide it completely

        Args:
            m: the map to display the layers
            kwargs: optional extra parameters for the ipyleaflet.WidgetControl
        """
        # save the map
        self.m = m

        # create a loading to place it on top of the card. It will always be visible
        # even when the card is scrolled
        p_style = json.loads((ss.JSON_DIR / "progress_bar.json").read_text())
        self.w_loading = sw.ProgressLinear(
            indeterminate=False,
            background_color=color.menu,
            color=p_style["color"][v.theme.dark],
        )
        self.tile = sw.Tile("nested", "")

        # set the kwargs parameters
        kwargs.setdefault("position", "topright")
        super().__init__(
            icon_content="fa-solid fa-layer-group",
            card_content=self.tile,
            m=m,
            **kwargs
        )

        # customize the menu to make it look more like a layercontrol
        self.menu.open_on_hover = True
        self.menu.close_delay = 200

        # update the table at instance creation
        self.update_table({})

        # add js behavior
        self.m.observe(self.update_table, "layers")

    def update_table(self, change: dict) -> None:
        """
        Update the table content.
        """
        # create the vector line
        vectors = [lyr for lyr in reversed(self.m.layers) if isinstance(lyr, GeoJSON)]
        vector_rows = []
        if len(vectors) > 0:
            head = [HeaderRow(ms.layer_control.vector.header)]
            rows = [VectorRow(lyr) for lyr in vectors]
            vector_rows = head + rows

        # create a table of layerLine
        layers = [
            lyr
            for lyr in reversed(self.m.layers)
            if lyr.base is False and isinstance(lyr, TileLayer)
        ]
        layer_rows = []
        if len(layers) > 0:
            head = [HeaderRow(ms.layer_control.layer.header)]
            rows = [LayerRow(lyr) for lyr in layers]
            layer_rows = head + rows

        # create another table of basemapLine it should always be a basemap
        # the error raised if you delete the last one is a feature
        bases = [lyr for lyr in self.m.layers if lyr.base is True]
        base_rows = []
        current = next(
            (lyr for lyr in bases if lyr.visible is True), SimpleNamespace(name=None)
        )
        if len(bases) > 0:
            head = [HeaderRow(ms.layer_control.basemap.header)]
            empy_cell = sw.Html(tag="td", children=[" "], attributes={"colspan": 3})
            empty_row = sw.Html(tag="tr", class_="v-no-hever", children=[empy_cell])
            rows = [BaseRow(lyr) for lyr in bases] + [empty_row]
            base_rows = head + rows

        # create a table from these rows and wrap it in the radioGroup
        tbody = sw.Html(tag="tbody", children=vector_rows + layer_rows + base_rows)
        table = sw.SimpleTable(children=[tbody], dense=True, class_="v-no-border")
        self.group = sw.RadioGroup(v_model=current.name, children=[table])

        # set the table as children of the widget
        self.tile.children = [self.group]

        return



"""
All modules used to create a application ``Map``.

The ``SepalMap`` is inheriting from the **ipyleaflet** ``Map``. It has been customized to support additional widgets as controls based on **ipyvuetify**. It is also fully compatible with the **sepal-ui** framework thanks to frontend modifications.

The main object is the ``SepalMap``that should be used in favor of ``Map`` in **sepal-ui** framework. The others are predisgned controls that can be helpful to speed up developments.
``SepalMap`` is fully compatible with Google Earth Engine layers.

Every module content can be called directly from the package.

Example:
    .. jupyter-execute::

        from sepal_ui import mapping as sm

        sm.SepalMap(gee=False)
"""

from .aoi_control import *
from .draw_control import *
from .fullscreen_control import *
from .layer import *
from .layer_state_control import *
from .layers_control import *
from .legend_control import *
from .map_btn import *
from .menu_control import *
from .sepal_map import *
from .value_inspector import *
from .zoom_control import *



"""
Custom sepalwidgets to add extra members to normal ``IpyvuetfiWidget``.

Gather the customized ``ipyvuetifyWidgets`` used toadd extra members to the exisiting one.
All the content of this modules is included in the parent ``sepal_ui.sepalwidgets`` package. So it can be imported directly from there.

Example:
    .. jupyter-execute::

        from sepal_ui import sepalwidgets as sw

        sw.Tooltip(widget=sw.Btn(), tooltip="tooltip")
"""

import warnings
from typing import Optional, Union

import ipyvuetify as v
import traitlets as t
from deprecated.sphinx import versionadded
from traitlets import observe
from typing_extensions import Self, Type

__all__ = ["SepalWidget", "Tooltip"]


class SepalWidget(v.VuetifyWidget):

    viz: t.Bool = t.Bool(True).tag(sync=True)
    "whether the widget is displayed or not"

    old_class: t.Unicode = t.Unicode("").tag(sync=True)
    "a saving attribute of the widget class"

    with_tooltip: Optional[v.Tooltip] = None
    "the full widget and its tooltip. Useful for display purposes when a tooltip has been set"

    def __init__(self, viz: bool = True, tooltip: str = "", **kwargs) -> None:
        """
        Custom vuetifyWidget to add specific methods.

        Args:
            viz: define if the widget should be visible or not
            tooltip: tooltip text of the widget. If set the widget will be displayed on :code:`self.widget` (irreversible).
        """
        # init the widget
        super().__init__(**kwargs)

        # setup the viz status
        self.viz = viz

        # create a tooltip only if a message is set
        self.set_tooltip(tooltip)

    @observe("viz")
    def _set_viz(self, *args) -> None:
        """
        Hide or show the component according to its viz param value.

        Hide the widget by reducing the html class to :code:`d-none`.
        Show the widget by removing the :code:`d-none` html class.
        Save the previous class
        """
        # will be replaced byt direct calls to built-in hide
        # once the previous custom implementation will be fully removed

        if self.viz:

            # change class value
            self.class_ = self.old_class or self.class_
            self.class_list.remove("d-none")

        else:

            # change class value
            self.class_list.remove("d-none")
            self.old_class = str(self.class_)
            self.class_ = "d-none"

        return

    def toggle_viz(self) -> Self:
        """Toogle the visibility of the widget."""
        self.viz = not self.viz

        return self

    def hide(self) -> Self:
        """
        Hide the widget by reducing the html class to :code:`d-none`.

        Save the previous class and set viz attribute to False.
        """
        # update viz state
        self.viz = False

        return self

    def show(self) -> Self:
        """
        Show the widget by removing the d-none html class.

        Save the previous class and set viz attribute to True.
        """
        # update viz state
        self.viz = True

        return self

    def reset(self) -> Self:
        """
        Clear the widget v_model.

        Need to be extented in custom widgets to fit the structure of the actual input.
        """
        self.v_model = None

        return self

    def get_children(
        self,
        widget: Optional[v.VuetifyWidget] = None,
        klass: Type[v.VuetifyWidget] = v.VuetifyWidget,
        attr: str = "",
        value: str = "",
        id_: str = "",
        elements: Optional[list] = None,
    ) -> list:
        r"""
        Recursively search for every element matching the specifications.

        multiple parameters can be used to search for matching elements. no error is raised if nothing is found.

        Args:
            widget: the widget to search into
            klass: the vuetify widget class to look for . Leave empty for any.
            attr: the attribute to look at. leave empty for no search
            value: the value of the attr. ignored if attr is not set
            elements: the list used to store found elements

        Returns:
            List containing all matching elements

        Retrieve all children elements that matches with the given id\_.

        Args:
            id\_ (str, optional): attribute id to compare with.

        Returns:
            list with all mathing elements if there are more than one, otherwise will return the matching element.

        """
        # id_ was the previous variable it should continue working in this implementation
        # remove kwargs when this will be deprecated
        if id_ != "":
            attr, value = ("id", id_)
            warnings.warn(
                '"id_" is a deprecated argument, use an ("attr", "value") pair instead',
                DeprecationWarning,
            )

        # init the element list
        elements = [] if elements is None else elements

        # init the widget
        widget = self if widget is None else widget

        for w in widget.children:

            # exit if children is not a widget (str, DOM objects.. etc)
            if not isinstance(w, v.VuetifyWidget):
                continue

            # compare the widget with requirements
            # using "niet" as default so that result is True if attr is Falsy
            # "niet" is very unlikely to be used compared to None, False, "none"...
            is_klass = isinstance(w, klass)
            is_val = w.attributes.get(attr, "niet") == value if attr and value else True

            # asumption: searched element won't be nested inside one another
            if is_klass and is_val:
                elements.append(w)
            else:
                elements = self.get_children(w, klass, attr, value, id_, elements)

        return elements

    def set_children(
        self, children: Union[str, v.VuetifyWidget, list], position: str = "first"
    ) -> Self:
        """
        Insert input children in self children within given position.

        Args:
            children: the list of children to add to the widget. It can also be a list (str and DOMWidgets are accepted)
            position: whether to insert as first or last element. ["first", "last"]
        """
        if not isinstance(children, list):
            children = [children]

        new_childrens = self.children[:]

        if position == "first":
            new_childrens = children + new_childrens

        elif position == "last":
            new_childrens = new_childrens + children

        else:
            raise ValueError(
                f"Position '{position}' is not a valid value. Use 'first' or 'last'"
            )

        self.children = new_childrens

        return self

    @versionadded(version="2.9.0", reason="Tooltip are now integrated to widgets")
    def set_tooltip(self, txt: str = "", **kwargs) -> v.Tooltip:
        """
        Create a tooltip associated with the widget.

        If the text is not set, the tooltip will be automatically removed. Once the tooltip is set the object variable can be accessed normally but to render the widget, one will need to use :code:`self.with_tooltip` (irreversible).

        Args:
            txt: anything False (0, False, empty text, None) will lead to the removal of the tooltip. everything else will be used to fill the text area
            kwargs: any options available in a Tooltip widget

        Returns:
            the tooltip associated with the object
        """
        if isinstance(self.with_tooltip, Tooltip):
            # If it's already created, and there are new kwargs, let's modify it
            [setattr(self.with_tooltip, attr, value) for attr, value in kwargs.items()]
            self.with_tooltip.children = [txt]
            self.with_tooltip.disabled = not bool(txt)
        elif txt != "":
            self.with_tooltip = Tooltip(self, txt, **kwargs)

        return self


class Tooltip(v.Tooltip):
    def __init__(self, widget: v.VuetifyWidget, tooltip: str, **kwargs) -> None:
        """
        Custom widget to display tooltip when mouse is over widget.

        Args:
            widget: widget used to display tooltip
            tooltip: the text to display in the tooltip
        """
        # set some default parameters
        kwargs.setdefault("close_delay", 200)

        self.v_slots = [
            {"name": "activator", "variable": "tooltip", "children": widget}
        ]
        widget.v_on = "tooltip.on"

        self.children = [tooltip]

        super().__init__(**kwargs)



"""
Customized control to toogle the fullscreen state of the map.
"""

from pathlib import Path
from typing import List, Optional

import ipyvuetify as v
from ipyleaflet import Map, WidgetControl
from IPython.display import Javascript, display

from sepal_ui.mapping.map_btn import MapBtn


class FullScreenControl(WidgetControl):

    ICONS: List[str] = ["fa-solid fa-expand", "fa-solid fa-compress"]
    "list: The icons that will be used to toggle between expand and compressed mode"

    METHODS: List[str] = ["embed", "fullscreen"]
    "list: The javascript methods name to be used to switch from expand to compress mode"

    zoomed: bool = False
    "bool: the current zoomed level: ``True`` for expanded and ``False`` for compressed"

    w_btn: Optional[v.Btn] = None
    "the btn to display on the map"

    template: Optional[v.VuetifyTemplate] = None
    "Embeds the 2 javascripts methods to change the rendering of the map"

    def __init__(
        self, m: Map, fullscreen: bool = False, fullapp: bool = False, **kwargs
    ) -> None:
        """
        A custom Fullscreen Button ready to be embed in a map object.

        This button will force the display of the map in fullscreen mode. It should be used instead of the built-in ipyleaflet FullscreenControl if your map is embeding ipyvuetify widgets. I tends to solve the issue raised here: https://github.com/widgetti/ipyvuetify/issues/141. The idea is to fake the fullscreen display by forcing the map container to extend to the full extend of the screen without using a z-index superior to the ipyvuetify overlay.
        simply click on it and the map will automatically expand

        .. versionadded:: 2.7.0

        Args:
            m: the map on which the mutated CSS will be applied (Only work with SepalMap as we are querying the _id)
            fullscreen: either the map should be displayed in fullscreen by default. default to false.
            fullapp: either or not the map will be used as the sole widget/tile of an application
            kwargs: any available arguments from a ipyleaflet WidgetControl
        """
        # set the offset
        offset = "48px" if fullapp else "0px"

        # register the required zoom value
        self.zoomed = fullscreen

        # create a btn
        self.w_btn = MapBtn(self.ICONS[self.zoomed])

        # overwrite the widget set in the kwargs (if any)
        kwargs["widget"] = self.w_btn
        kwargs.setdefault("position", "topleft")
        kwargs["transparent_bg"] = True

        # create the widget
        super().__init__(**kwargs)

        # add javascrip behaviour
        self.w_btn.on_event("click", self.toggle_fullscreen)

        # save the 2 fullscrenn js code in a table 0 for embeded and 1 for fullscreen
        js_dir = Path(__file__).parents[1] / "frontend/js"
        embed = (js_dir / "jupyter_embed.js").read_text() % m._id
        full = (js_dir / "jupyter_fullscreen.js").read_text() % (m._id, offset)

        # template with js behaviour
        # "jupyter_fullscreen" place tje "leaflet-container element on the front screen
        # and expand it's display to the full screen
        # "jupyter_embed" reset all the changed parameter
        # both trigger the resize event to force the reload of the Tilelayers
        self.template = v.VuetifyTemplate(
            template=(
                "<script>{methods: {jupyter_embed(){%s}, jupyter_fullscreen(){%s}}}</script>"
                % (embed, full)
            )
        )
        display(self.template)

        # display the map in the requested default state
        js = full if self.zoomed else embed
        display(Javascript(js))

    def toggle_fullscreen(self, *args) -> None:
        """
        Toogle fullscreen state.

        Toggle the fullscreen state of the map by sending the required javascript method,
        changing the w_btn icons and the zoomed state of the control.
        """
        # change the zoom state
        self.zoomed = not self.zoomed

        # change button icon
        self.w_btn.children[0].children = [self.ICONS[self.zoomed]]

        # zoom
        self.template.send({"method": self.METHODS[self.zoomed], "args": []})

        return



"""
Based ``SepalMap`` Btn.
"""

import ipyvuetify as v

from sepal_ui import color
from sepal_ui import sepalwidgets as sw


class MapBtn(v.Btn, sw.SepalWidget):
    def __init__(self, content: str, **kwargs) -> None:
        """
        Btn specifically design to be displayed on a map.

        It matches all the characteristics of
        the classic leaflet btn but as they are from ipyvuetify we can use them in combination with Menu to produce on-the-map tiles.
        The MapBtn is responsive to theme changes. It only accept icon or 3 letters as children as the space is very limited.

        Args:
            content: a fa-solid/mdi fully qualified name or a string name. If a string name is used, only the 3 first letters will be displayed.
        """
        # create the icon
        if content.startswith("mdi-") or content.startswith("fa"):
            content = sw.Icon(small=True, children=[content])
        else:
            content = content[: min(3, len(content))].upper()

        # some parameters are overloaded to match the map requirements
        kwargs["color"] = "text-color"
        kwargs["outlined"] = True
        kwargs["style_"] = f"background: {color.bg};"
        kwargs["children"] = [content]
        kwargs["icon"] = False
        kwargs.setdefault("class_", "v-map-btn")

        super().__init__(**kwargs)



#!/usr/bin/python3

"""
Script to update the requirements file with the currently used libs.

The script should be launched from a module directory.
It will parse all the files and extract the differnet librairies used in the module. They will be added to the requirements.txt
file using the versions used in the current local installation.
Some troubleshouting are handled by the script:
-  the earthengine-api will be forced to use the openforis fork to be compatible with SEPAL
-  pyproj and pygdal version will be forced to the version set in SEPAl as well
-  sepal-ui version will be bound to the one available when calling the script
Once the file have been created, please check it manually and make sure that there are no visible issues. PLease report any incompatibility to the developer
so taht they can be added to the troubleshoot function.
"""

import argparse
import subprocess
from pathlib import Path
from typing import Union

from colorama import Fore, Style, init

import sepal_ui

# init colors for all plateforms
init()

# init parser
parser = argparse.ArgumentParser(description=__doc__, usage="module_deploy")


def write_reqs(file: Union[str, Path]) -> None:
    """
    write the requirements in the requirements file.

    Args:
        file: the requirements file
    """
    # force cast to pathlib.Path
    file = Path(file)

    with file.open("a") as f:
        f.write("\n")
        f.write("\n# custom libs")
        f.write("\n")

    # add the custom libs
    tmp_file = Path.cwd() / "req_tmp.txt"
    subprocess.run(
        ["pipreqs", "--savepath", str(tmp_file), str(Path.cwd())], cwd=Path.cwd()
    )

    # add the libs in the final file
    with file.open("a") as dst:
        dst.write(tmp_file.read_text())

    # remove the tmp file
    tmp_file.unlink()

    return


def clean_dulpicate(file: Union[str, Path]) -> None:
    """
    Remove the requirements that are already part of the default installation.

    Args:
        file: the requirements file
    """
    # already available libs
    libs = ["wheel", "Cython", "pybind11", "GDAL", "pyproj", "sepal_ui"]

    file = Path(file)
    text = file.read_text().split("\n")

    # search for the custom line index
    idx = text.index("# custom libs")

    final_text = text[:idx]
    for line in text[idx:]:
        if any(lib in line for lib in libs):
            lib = next(lb for lb in libs if lb in line)
            print(
                f"Removing {Style.BRIGHT}{lib}{Style.NORMAL} from reqs, duplicated from default."
            )
            continue
        final_text.append(line)

    # write to file
    file.write_text("\n".join(final_text))

    return


def clean_troubleshouting(file: Union[str, Path]) -> None:
    """
    Remove know buggish lib from the requirements.

    The pipreqs is creating the file based on the import statements in .py files
    some libs doesn't have the same name as the pip command we are replacing/deleting the known one.

    Args:
        file: the requirements file
    """
    file = Path(file)
    text = file.read_text().split("\n")

    # search for the custom line index
    idx = text.index("# custom libs")

    final_text = text[:idx]
    for line in text[idx:]:

        # ee is imported as ee but the real name of the lib is earthengine-api
        # gdal and osgeo are part of pygdal
        # we use a specific version of earthengine in SEPAl, let's stick to it
        if "ee" in line:
            print(
                f"Removing {Style.BRIGHT}ee{Style.NORMAL} from reqs, included in sepal_ui."
            )
            continue
        elif any(lib in line for lib in ["osgeo", "gdal"]):
            print(
                f"Removing {Style.BRIGHT}'osgeo & gdal'{Style.NORMAL} from reqs, included in GDAL."
            )
            continue
        elif "earthengine_api" in line:
            print(
                f"Removing {Style.BRIGHT}earthengine_api{Style.NORMAL} from reqs, included in sepal_ui."
            )
            continue

        # if I'm here nothing needed to be removed
        final_text.append(line)

    file.write_text("\n".join(final_text))

    return


def freeze_sepal_ui(file: Union[str, Path]) -> None:
    """
    Set the sepal version to the currently used sepal-ui version.

    Args:
        file: the requirements file
    """
    file = Path(file)
    text = file.read_text().split("\n")

    # search for the sepal_ui line
    idx, _ = next(
        (i, il) for i, il in enumerate(text) if "#" not in il and "sepal_ui" in il
    )

    text[idx] = f"sepal_ui=={sepal_ui.__version__}"

    file.write_text("\n".join(text))

    print(
        f"sepal_ui version have been freezed to  {Style.BRIGHT}{sepal_ui.__version__}{Style.NORMAL}"
    )

    return


def clean_custom(file: Union[str, Path]) -> None:
    """
    Remove the previous custom installation and requirements.

    To start the process from a blank page.

    Args:
        file: the requirements file
    """
    file = Path(file)
    text = file.read_text().split("\n")

    # search for the custom line index
    try:
        idx = text.index("# custom libs")

        # remove everything after it
        file.write_text("\n".join(text[: idx - 1]))

        print("the previously set custom requirements have been cleaned from the file")

    except ValueError:
        pass

    return


def main() -> None:
    """Launch the process."""
    # parse agruments
    parser.parse_args()

    # welcome the user
    print(f"{Fore.YELLOW}sepal-ui module deployment tool{Fore.RESET}")

    print("Export the env configuration of your module...")

    # check that the local folder is a module folder
    ui_file = Path.cwd() / "ui.ipynb"
    if not ui_file.is_file():
        raise Exception(f"{Fore.RED}This is not a module folder.")

    # add the requirements to the requirements.txt file
    req_file = Path.cwd() / "requirements.txt"

    # clean the already existing custom installation
    # if it was done manually nothing will be removed
    clean_custom(req_file)

    # write the new requirements
    write_reqs(req_file)

    # clean the requirement file from known troubleshoutings
    clean_dulpicate(req_file)
    clean_troubleshouting(req_file)

    # set the sepal version
    freeze_sepal_ui(req_file)

    # exit message
    print(f'{Fore.GREEN}The "requirements.txt" file have been updated.{Fore.RESET}')
    print(
        f"{Fore.YELLOW}The tool does not cover every possible configuration so don't forget to check the final file before pushing to release.{Fore.RESET}"
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""
Script to manually change the used theme.

This script will update the parameters shared between all sepal-ui based modules.
The "language" parameter will be changed to the selected one. Any running application will need to be restarted to use this modification
"""

import argparse
from pathlib import Path

import pandas as pd
from colorama import Fore, init

from sepal_ui.scripts import utils as su

# init colors for all plateforms
init()

# init the parser
parser = argparse.ArgumentParser(description=__doc__, usage="module_l10n")


def check_locale(locale: str) -> bool:
    """
    Check if the locale exist in the country list.

    Args:
        locale: the locale name in iso code

    Returns:
        True if the language is a well defined
    """
    file = Path(__file__).parents[1] / "data" / "locale.parquet"
    countries = pd.read_parquet(file).astype(str)

    return locale in countries.code.values


def main() -> None:
    """Launch the process."""
    # parse agruments
    parser.parse_args()

    # welcome the user
    print(f"{Fore.YELLOW}sepal-ui localisation script{Fore.RESET}")

    # select a language
    is_locale = False
    while is_locale is False:

        locale = input(f"{Fore.CYAN}Provide a locale code: \n{Fore.RESET}")
        is_locale = check_locale(locale)

        # display an error if the language does not exist
        if is_locale is False:
            print(
                f'{Fore.RED} The provided language code ("{locale}") is not a valid language code in IETF BCP 47.{Fore.RESET}'
            )

    # write the new color code in the config file
    su.set_config_locale(locale)

    # display information
    print(
        f'{Fore.GREEN} The provided language code ("{locale}") has been set as default language for all SEPAL applications.{Fore.RESET}'
    )

    return


if __name__ == "__main__":
    main()



"""
The ``Card`` widget to use in application to interface with Planet.
"""

from typing import Optional

import ipyvuetify as v

import sepal_ui.sepalwidgets as sw
from sepal_ui.message import ms
from sepal_ui.planetapi.planet_model import PlanetModel
from sepal_ui.planetapi.planet_widgets import InfoView
from sepal_ui.scripts.decorator import loading_button


class PlanetView(sw.Layout):

    planet_model: Optional[PlanetModel] = None
    "Backend model to manipulate interface actions"

    btn: Optional[sw.Btn] = None
    "Button to trigger the validation process in the associated model"

    alert: Optional[sw.Alert] = None
    "Alert component to display end-user action results"

    info: bool = False
    "either to display or not a detailed description about the planet subscriptions"

    w_username: Optional[sw.TextField] = None
    "Widget to set credential username"

    w_password: Optional[sw.PasswordField] = None
    "Widget to set credential password"

    w_key: Optional[sw.PasswordField] = None
    "Widget to set credential API key"

    w_method: Optional[sw.Select] = None
    "Dropdown widget to select connection method"

    def __init__(
        self,
        btn: Optional[sw.Btn] = None,
        alert: Optional[sw.Alert] = None,
        planet_model: Optional[PlanetModel] = None,
        info: bool = False,
        **kwargs,
    ):
        """
        Stand-alone interface to capture planet lab credentials.

        It also validate its  subscription and connect to the client stored in the model.

        Args:
            btn (sw.Btn, optional): Button to trigger the validation process in the associated model.
            alert (sw.Alert, v.Alert, optional): Alert component to display end-user action results.
            planet_model (sepal_ui.planetlab.PlanetModel): backend model to manipulate interface actions.
        """
        self.class_ = "d-block flex-wrap"

        super().__init__(**kwargs)

        self.planet_model = planet_model if planet_model else PlanetModel()
        self.btn = btn if btn else sw.Btn("Validate", small=True, class_="mr-1")
        self.alert = alert if alert else sw.Alert()

        self.w_username = sw.TextField(
            label=ms.planet.widget.username, class_="mr-2", v_model=""
        )
        self.w_password = sw.PasswordField(label=ms.planet.widget.password)
        self.w_key = sw.PasswordField(label=ms.planet.widget.apikey, v_model="").hide()
        self.w_info_view = InfoView(model=self.planet_model)

        self.w_method = v.Select(
            label=ms.planet.widget.method.label,
            class_="mr-2",
            v_model="credentials",
            items=[
                {"value": "credentials", "text": ms.planet.widget.method.credentials},
                {"value": "api_key", "text": ms.planet.widget.method.api_key},
            ],
        )

        w_validation = v.Flex(
            style_="flex-grow: 0 !important;",
            children=[self.btn],
            class_="pr-1 flex-nowrap",
        )
        self.children = [
            self.w_method,
            sw.Layout(
                class_="align-center",
                children=[
                    self.w_username,
                    self.w_password,
                    self.w_key,
                ],
            ),
        ]

        if not btn:
            self.children[-1].set_children(w_validation, "last")

        # Set it here to avoid displacements when using button
        self.set_children(self.w_info_view, "last")

        if not alert:
            self.set_children(self.alert, "last")

        self.w_method.observe(self._swap_inputs, "v_model")
        self.btn.on_event("click", self.validate)

    def reset(self) -> None:
        """Empty credentials fields and restart activation mode."""
        self.w_username.v_model = None
        self.w_password.v_model = None
        self.w_key.v_model = None
        self.planet_model.__init__()

        return

    def _swap_inputs(self, change: dict) -> None:
        """Swap between credentials and api key inputs."""
        self.alert.reset()
        self.reset()

        self.w_username.toggle_viz()
        self.w_password.toggle_viz()
        self.w_key.toggle_viz()

        return

    @loading_button(debug=True)
    def validate(self, *args) -> None:
        """Initialize planet client and validate if is active."""
        self.planet_model.__init__()

        if self.w_method.v_model == "credentials":
            credentials = [self.w_username.v_model, self.w_password.v_model]
        else:
            credentials = [self.w_key.v_model]

        self.planet_model.init_session(credentials)

        return



"""
A specific statebar dedicated to the the counting of loading tiles in the map.
"""

from typing import Optional

import traitlets as t
from ipyleaflet import Map, WidgetControl
from traitlets import observe

from sepal_ui import sepalwidgets as sw
from sepal_ui.message import ms


class LayerStateControl(WidgetControl):

    m: Optional[Map] = None
    "the map connected to the control"

    w_state: Optional[sw.StateBar] = None
    "sw.StateBar: the stateBar displaying the number of layer loading on the map"

    nb_layer: t.Int = t.Int(0).tag(sync=True)
    "the number of layers in the map"

    nb_loading_layer: t.Int = t.Int(0).tag(sync=True)
    "the number of loading layer in the map"

    def __init__(self, m: Map, **kwargs) -> None:
        """
        A specific statebar dedicated to the the counting of loading tiles in the map.

        every time a map is added to the map the counter will be raised by one. same behaviour with removed.

        Args:
            m: The map the component is listening to
        """
        # save the map as a member of the widget
        self.m = m

        # create a statebar
        msg = ms.layer_state.complete.format(self.nb_layer)
        self.w_state = sw.StateBar(loading=False, msg=msg)

        # overwrite the widget set in the kwargs (if any)
        kwargs["widget"] = self.w_state
        kwargs.setdefault("position", "topleft")
        kwargs["transparent_bg"] = True

        # create the widget
        super().__init__(**kwargs)

        # add js behaviour
        self.m.observe(self.update_nb_layer, "layers")

    def update_nb_layer(self, change: dict) -> None:
        """
        Update the number of layer monitored by the statebar.
        """
        # exit if nothing changed
        # for example we change a layer parameters and it trigger this one
        if len(change["new"]) == len(change["old"]):
            return

        self.nb_layer = len([lyr for lyr in change["new"] if not lyr.base])

        # identify the modified layer
        modified_layer = list(set(change["new"]) ^ set(change["old"]))[0]
        if modified_layer.base is True:
            return

        # add a layer
        if len(change["new"]) > len(change["old"]):
            modified_layer.observe(self.update_loading, "loading")

        # remove a layer
        elif len(change["new"]) < len(change["old"]):
            # the test is splitted as not all the layers have a loading trait
            if hasattr(modified_layer, "loading") is True:
                if modified_layer.loading is True:
                    self.nb_loading_layer += -1

        return

    def update_loading(self, change: dict) -> None:
        """update the nb_loading_layer value according to the number of tile loading on the map."""
        increment = [-1, 1]
        self.nb_loading_layer += increment[change["new"]]

        return

    @observe("nb_loading_layer", "nb_layer")
    def _update_state(self, *args) -> None:

        # check if anything is loading
        self.loading = bool(self.nb_loading_layer)

        # update the message
        if self.loading is True:
            msg = ms.layer_state.loading.format(self.nb_loading_layer, self.nb_layer)
        else:
            msg = ms.layer_state.complete.format(self.nb_layer)

        self.w_state.msg = msg

        return



"""
ResizeTrigger is an essential object of sepal-ui.

It guarantees that the maps displayed in the application are allways resized when you change the view. Without it moving from one panel to another would make some of the map tile invisible.
"""

from pathlib import Path

import ipyvuetify as v
from IPython.display import display
from traitlets import Unicode


class ResizeTrigger(v.VuetifyTemplate):
    """
    A trigger to resize maps when a change of display is done.

    Every time resize is called, the javascript resize event is trigger of the application.
    """

    # load the js file
    js = (Path(__file__).parent / "js/jupyter_resize.js").read_text()
    template = Unicode("<script>{methods: {jupyter_resize(){%s}}}</script>" % js).tag(
        sync=True
    )
    "Unicode: the javascript script to manually trigger the resize event"

    def resize(self):
        """trigger the template method i.e. the resize event."""
        return self.send({"method": "resize"})


# create one single resizetrigger that will be used as a singleton everywhere
# singletons are bad but if we display multiple instances of rt for every DrawItem
# the initial offset will be impossible to manage
rt = ResizeTrigger()
display(rt)



"""
Based menu ``Control`` to display widgets to your user.
"""

from typing import Optional, Union

import ipyvuetify as v
from ipyleaflet import Map, WidgetControl
from traitlets import Int
from typing_extensions import Self

from sepal_ui import sepalwidgets as sw
from sepal_ui.mapping.map_btn import MapBtn


class MenuControl(WidgetControl):

    menu: Optional[sw.Menu] = None
    "the menu displayed on the map as a widget"

    m: Optional[Map] = None
    "the map used to display the control"

    group: Int = Int().tag(sync=True)
    "The group of Menu control, tell how the Menucontrol interact with the others"

    def __init__(
        self,
        icon_content: str,
        card_content: Union[v.VuetifyWidget, str],
        card_title: str = "",
        m: Optional[Map] = None,
        group: int = 0,
        **kwargs
    ) -> None:
        """
        Widget control displaying a btn on the map.

        When clicked the menu expand to show the content set by the user and all the others are closed.
        It's used to display interactive tiles directly in the map. If the card_content is a Tile it will be automatically nested.

        Args:
            icon_content: the icon content as specified in the sm.MapBtn object (i.e. a 3 letter name or an icon name)
            card_content: any container from sw. The sw.Tile is specifically design to fit in this component
            card_title: the card title. THe tile title will override this parameter if existing
            m: The map associated with the Menu
            group: The group of Menu control, tell how the Menucontrol interact with the others
        """
        # save the map in the members
        self.m = m

        # set the menucontrol group
        self.group = group

        # create a clickable btn
        btn = MapBtn(content=icon_content, v_on="menu.on")
        slot = {"name": "activator", "variable": "menu", "children": btn}

        # nest the content if it's a sw.Tile
        children = [card_content]
        if isinstance(card_content, sw.Tile):
            card_title = card_content.get_title()
            card_content.nest()
            card_content.class_list.replace("ma-5", "ma-0")
            card_content.children[0].class_list.replace("pa-5", "pa-2")
            card_content.children[0].raised = False
            card_content.children[0].elevation = 0

        # set up a title of needed
        if card_title != "":
            card_title = sw.CardTitle(children=[card_title])
            children.insert(0, card_title)

        # set up the content style
        card = sw.Card(
            tile=True,
            style_="overflow: auto",
            children=children,
        )

        # assemble everything in a menu
        self.menu = sw.Menu(
            v_model=False,
            close_on_click=False,
            close_on_content_click=False,
            children=[card],
            v_slots=[slot],
            offset_x=True,
        )

        # by default MenuControls will be displayed in the bottomright corner
        kwargs.setdefault("position", "bottomright")
        kwargs["widget"] = self.menu

        super().__init__(**kwargs)

        # place te menu according to the widget positioning
        self.update_position(None)
        self.set_size()

        # add some interaction
        self.observe(self.update_position, "position")
        self.menu.observe(self.close_others, "v_model")

    def update_position(self, *args) -> None:
        """
        Update the position of the menu if the position of the widget is dynamically changed.
        """
        self.menu.top = "bottom" in self.position
        self.menu.bottom = "top" in self.position
        self.menu.left = "right" in self.position
        self.menu.right = "left" in self.position

        return

    def set_size(
        self,
        min_width: Optional[Union[str, int]] = "400px",
        max_width: Optional[Union[str, int]] = "400px",
        min_height: Optional[Union[str, int]] = "40vh",
        max_height: Optional[Union[str, int]] = "40vh",
    ) -> Self:
        """
        Set the size of the card using all the sizing parameters from a v.Card.

        Args:
          min_width: a fully qualified css description of the wanted min_width. default to 400px.
          max_width: a fully qualified css description of the wanted max_width. default to 400px.
          min_height: a fully qualified css description of the wanted min_height. default to 40vh.
          max_height: a fully qualified css description of the wanted max_height. default to 40vh.
        """
        card = self.menu.children[0]

        card.min_width = min_width
        card.max_width = max_width
        card.min_height = min_height
        card.max_height = max_height

        return self

    def close_others(self, *args) -> None:
        """
        Close all the other menus associated to the map to avoid overlapping.

        all the other Menu control from the same group will be closed if this menu is opened. Other MenuCOntrol from other groups can remain open. User will need to be careful with Widget placements on the map.
        """
        # don't do anything if no map was set to avoid deprecation
        # remove when jumping to sepal-ui 3.0
        if self.m is None:
            return

        # avoid infinite loop by exiting the method when it's closed
        if self.menu.v_model is True:
            [
                setattr(c.menu, "v_model", False)
                for c in self.m.controls
                if isinstance(c, MenuControl) and c != self and c.group == self.group
            ]

        return



"""
Module to load basemaps from different providers.
"""

from typing import Optional

from box import Box
from ipyleaflet import TileLayer
from xyzservices import TileProvider
from xyzservices import providers as xyz

xyz_tiles: dict = {
    "OpenStreetMap": {
        "url": "https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png",
        "attribution": "OpenStreetMap",
        "name": "OpenStreetMap",
    },
    "ROADMAP": {
        "url": "https://mt1.google.com/vt/lyrs=m&x={x}&y={y}&z={z}",
        "attribution": "Google",
        "name": "Google Maps",
    },
    "SATELLITE": {
        "url": "https://mt1.google.com/vt/lyrs=s&x={x}&y={y}&z={z}",
        "attribution": "Google",
        "name": "Google Satellite",
    },
    "TERRAIN": {
        "url": "https://mt1.google.com/vt/lyrs=p&x={x}&y={y}&z={z}",
        "attribution": "Google",
        "name": "Google Terrain",
    },
    "HYBRID": {
        "url": "https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}",
        "attribution": "Google",
        "name": "Google Satellite",
    },
}
"Custom XYZ tile services."


def get_xyz_dict(
    free_only: bool = True,
    _collection: Optional[dict] = None,
    _output: Optional[dict] = None,
) -> dict:
    """
    Returns a dictionary of xyz services.

    Args:
        free_only: Whether to return only free xyz tile services that do not require an access token.
        _collection: the collection to anylize (subset of :code:`xyz`)
        _output: the dict to use as an output (mutable object)

    Returns:
        A dictionary of xyz services.
    """
    # the 2 following lies avoid to display xyz descriptor in the method documentation
    # do not replace in the prototype default values
    _collection = xyz if _collection is None else _collection
    _output = {} if _output is None else _output

    for v in _collection.values():
        if isinstance(v, TileProvider):
            if not (v.requires_token() and free_only):
                _output[v.name] = v
        else:  # it's a Bunch
            get_xyz_dict(free_only, v, _output)

    return _output


def xyz_to_leaflet() -> dict:
    """
    Convert all available xyz tile services to ipyleaflet tile layers.

    Adapted from https://github.com/giswqs/geemap.

    Returns:
        A dictionary of ipyleaflet tile layers.
    """
    leaflet_dict = {}

    for key in xyz_tiles:
        name = xyz_tiles[key]["name"]
        url = xyz_tiles[key]["url"]
        attribution = xyz_tiles[key]["attribution"]
        leaflet_dict[key] = TileLayer(
            url=url, name=name, attribution=attribution, max_zoom=22, base=True
        )

    for item in get_xyz_dict().values():
        leaflet_dict[item.name] = TileLayer(
            url=item.build_url(),
            name=item.name,
            max_zoom=item.get("max_zoom", 22),
            attribution=item.attribution,
            base=True,
        )

    return leaflet_dict


basemap_tiles: Box = Box(xyz_to_leaflet(), frozen_box=True)
"the basemaps list as a box"



"""
Custom widgets relative to user application framework.

Gather the customized ``ipyvuetifyWidgets`` used to create the application framework.
All the content of this modules is included in the parent ``sepal_ui.sepalwidgets`` package. So it can be imported directly from there.

Example:
    .. jupyter-execute::
    
        from sepal_ui import sepalwidgets as sw
        
        sw.LocaleSelect()
"""

from datetime import datetime
from functools import partial
from itertools import cycle
from pathlib import Path
from typing import Dict, List, Optional, Union

import ipyvuetify as v
import pandas as pd
import traitlets as t
from deprecated.sphinx import versionadded, versionchanged
from ipywidgets import jsdlink
from traitlets import link, observe
from typing_extensions import Self

from sepal_ui import color
from sepal_ui.frontend.resize_trigger import ResizeTrigger, rt
from sepal_ui.frontend.styles import get_theme
from sepal_ui.message import ms
from sepal_ui.model import Model
from sepal_ui.scripts import utils as su
from sepal_ui.sepalwidgets.alert import Banner
from sepal_ui.sepalwidgets.sepalwidget import SepalWidget
from sepal_ui.translator import Translator

__all__ = [
    "AppBar",
    "DrawerItem",
    "NavDrawer",
    "Footer",
    "App",
    "LocaleSelect",
    "ThemeSelect",
]


class LocaleSelect(v.Menu, SepalWidget):

    COUNTRIES: pd.DataFrame = pd.read_parquet(
        Path(__file__).parents[1] / "data" / "locale.parquet"
    )
    "the country list as a df. columns [code, name, flag]"

    FLAG: str = "https://flagcdn.com/{}.svg"
    "the url of the svg flag images"

    ATTR: Dict[str, str] = {
        "src": "https://flagcdn.com/gb.svg",
        "width": "30",
        "alt": "en-UK",
    }
    "the default flag parameter, default to english"

    btn: Optional[v.Btn] = None
    "the btn to click when changing language"

    language_list: Optional[v.List] = None
    "the list of countries with their flag,name in english, and ISO code"

    def __init__(self, translator: Optional[Translator] = None, **kwargs) -> None:
        """
        A language selector for sepal-ui based application.

        It displays the currently requested language (not the one used by the translator).
        When value is changed, the sepal-ui config file is updated. It is designed to be used in a AppBar component.

        .. warning:: as the component is a v.Menu to get the selected value you need to lisen to "value" instead of "v_model".

        .. versionadded:: 2.7.0

        Args:
            translator: the translator of the app, to match the used language
            kwargs (optional): any arguments for a Btn object, children will be override
        """
        # extract the available language from the translator
        # default to only en-US if no translator is set
        available_locales = (
            ["en"] if translator is None else translator.available_locales()
        )

        # extract the language information from the translator
        # if not set default to english
        code = "en" if translator is None else translator._target
        loc = self.COUNTRIES[self.COUNTRIES.code == code].squeeze()
        attr = {**self.ATTR, "src": self.FLAG.format(loc.flag), "alt": loc.name}

        kwargs.setdefault("small", True)
        kwargs["v_model"] = False
        kwargs["v_on"] = "x.on"
        kwargs["children"] = [v.Html(tag="img", attributes=attr, class_="mr-1"), code]
        self.btn = v.Btn(**kwargs)

        self.language_list = v.List(
            dense=True,
            flat=True,
            color=color.menu,
            v_model=True,
            max_height="300px",
            style_="overflow: auto; border-radius: 0 0 0 0;",
            children=[
                v.ListItemGroup(
                    children=self._get_country_items(available_locales), v_model=""
                )
            ],
        )

        super().__init__(
            children=[self.language_list],
            v_model=False,
            close_on_content_click=True,
            v_slots=[{"name": "activator", "variable": "x", "children": self.btn}],
            value=loc.code,
        )

        # add js behaviour
        jsdlink((self.language_list.children[0], "v_model"), (self, "value"))
        self.language_list.children[0].observe(self._on_locale_select, "v_model")

    def _get_country_items(self, locales: list) -> List[str]:
        """
        Get the list of countries as a list of listItem.

        Reduce the list to the available language of the module.

        Args:
            locales: list of the locales to display

        Returns:
            the list of contry widget to display in the app
        """
        country_list = []
        filtered_countries = self.COUNTRIES[self.COUNTRIES.code.isin(locales)]
        for r in filtered_countries.itertuples(index=False):

            attr = {**self.ATTR, "src": self.FLAG.format(r.flag), "alt": r.name}

            children = [
                v.ListItemAction(children=[v.Html(tag="img", attributes=attr)]),
                v.ListItemContent(children=[v.ListItemTitle(children=[r.name])]),
                v.ListItemActionText(children=[r.code]),
            ]

            country_list.append(v.ListItem(value=r.code, children=children))

        return country_list

    def _on_locale_select(self, change: dict) -> None:
        """
        adapt the application to the newly selected language.

        Display the new flag and country code on the widget btn
        change the value in the config file
        """
        # get the line in the locale dataframe
        loc = self.COUNTRIES[self.COUNTRIES.code == change["new"]].squeeze()

        # change the btn attributes
        attr = {**self.ATTR, "src": self.FLAG.format(loc.flag), "alt": loc.name}
        self.btn.children = [
            v.Html(tag="img", attributes=attr, class_="mr-1"),
            loc.code,
        ]
        self.btn.color = "info"

        # change the paramater file
        su.set_config("locale", loc.code)

        return


class ThemeSelect(v.Btn, SepalWidget):

    THEME_ICONS: dict = {"dark": "fa-solid fa-moon", "light": "fa-solid fa-sun"}
    "the dictionnry of icons to use for each theme (used as keys)"

    theme: str = "dark"
    "the current theme of the widget (default to dark)"

    def __init__(self, **kwargs) -> None:
        """
        A theme selector for sepal-ui based application.

        It displays the currently requested theme (default to dark).
        When value is changed, the sepal-ui config file is updated. It is designed to be used in a AppBar component.

        .. versionadded:: 2.7.0

        Args:
            kwargs: any arguments for a Btn object, children and v_model will be override
        """
        # get the current theme name
        self.theme = get_theme()

        # set the btn parameters
        kwargs.setdefault("x_small", True)
        kwargs.setdefault("fab", True)
        kwargs.setdefault("class_", "ml-2")
        kwargs["children"] = [v.Icon(children=[self.THEME_ICONS[self.theme]])]
        kwargs["v_model"] = self.theme

        # create the btn
        super().__init__(**kwargs)

        # add some js events
        self.on_event("click", self.toggle_theme)

    def toggle_theme(self, *args) -> None:
        """
        Toggle the btn icon from dark to light and adapt the configuration file.
        """
        # use a cycle to go through the themes
        theme_cycle = cycle(self.THEME_ICONS.keys())
        next(t for t in theme_cycle if t == self.theme)
        self.theme = next(t for t in theme_cycle)

        # change icon
        self.color = "info"
        self.children[0].children = [self.THEME_ICONS[self.theme]]

        # change the paramater file
        su.set_config("theme", self.theme)

        # trigger other events by changing v_model
        self.v_model = self.theme

        return


class AppBar(v.AppBar, SepalWidget):

    toogle_button: Optional[v.Btn]
    "The btn to display or hide the drawer to the user"

    title: Optional[v.ToolbarTitle]
    "The widget containing the app title"

    locale: Optional[LocaleSelect]
    "The locale selector of all apps"

    theme = Optional[ThemeSelect]
    "The theme selector of all apps"

    def __init__(
        self,
        title: str = "SEPAL module",
        translator: Union[None, Translator] = None,
        **kwargs,
    ) -> None:
        """
        Custom AppBar widget with the provided title using the sepal color framework.

        Args:
            title: the title of the app
            translator: the app translator to pass to the locale selector object
            kwargs (optional): any parameters from a v.AppBar. If set, 'children' and 'app' will be overwritten.
        """
        self.toggle_button = v.Btn(
            icon=True,
            children=[
                v.Icon(class_="white--text", children=["fa-solid fa-ellipsis-v"])
            ],
        )

        self.title = v.ToolbarTitle(children=[title])

        self.locale = LocaleSelect(translator=translator)
        self.theme = ThemeSelect()

        # set the default parameters
        kwargs.setdefault("color", color.main)
        kwargs.setdefault("class_", "white--text")
        kwargs.setdefault("dense", True)
        kwargs["app"] = True
        kwargs["children"] = [
            self.toggle_button,
            self.title,
            v.Spacer(),
            self.locale,
            self.theme,
        ]

        super().__init__(**kwargs)

    def set_title(self, title: str) -> Self:
        """
        Set the title of the appBar.

        Args:
            title: the new app title
        """
        self.title.children = [title]

        return self


class DrawerItem(v.ListItem, SepalWidget):

    rt: Optional[ResizeTrigger] = None
    "The trigger to resize maps and other javascript object when jumping from a tile to another"

    alert: t.Bool = t.Bool(False).tag(sync=True)
    "Trait to control visibility of an alert in the drawer item"

    alert_badge: Optional[v.ListItemAction]
    "red circle to display in the drawer"

    def __init__(
        self,
        title: str,
        icon: str = "",
        card: str = "",
        href: str = "",
        model: Union[Model, None] = None,
        bind_var: str = "",
        **kwargs,
    ) -> None:
        """
        Custom DrawerItem using the user input.

        If a card is set the drawerItem will trigger the display of all the Tiles in the app that have the same mount_id.
        If an href is set, the drawer will open the link in a new tab.

        Args:
            title: the title of the drawer item
            icon: the full name of a mdi/fa icon
            card: the mount_id of tiles in the app
            href: the absolute link to an external web page
            model: sepalwidget model where is defined the bin_var trait
            bind_var: required when model is selected. Trait to link with 'alert' self trait parameter
            kwargs (optional): any parameter from a v.ListItem. If set, '_metadata', 'target', 'link' and 'children' will be overwritten.
        """
        # set the resizetrigger
        self.rt = rt

        icon = icon if icon else "fa-regular fa-folder"

        children = [
            v.ListItemAction(children=[v.Icon(class_="white--text", children=[icon])]),
            v.ListItemContent(
                children=[v.ListItemTitle(class_="white--text", children=[title])]
            ),
        ]

        # set default parameters
        kwargs["link"] = True
        kwargs["children"] = children
        kwargs.setdefault("input_value", False)
        if href:
            kwargs["href"] = href  # cannot be set twice anyway
            kwargs["target"] = "_blank"
            kwargs.setdefault("_metadata", None)
        elif card:
            kwargs["_metadata"] = {"card_id": card}
            kwargs.setdefault("href", None)
            kwargs.setdefault("target", None)

        # call the constructor
        super().__init__(**kwargs)

        # cannot be set as a class member because it will be shared with all
        # the other draweritems.
        self.alert_badge = v.ListItemAction(
            children=[
                v.Icon(children=["fa-solid fa-circle"], x_small=True, color="red")
            ]
        )

        if model:
            if not bind_var:
                raise Exception(
                    "You have selected a model, you need a trait to bind with drawer."
                )

            link((model, bind_var), (self, "alert"))

    @observe("alert")
    def add_notif(self, change: dict) -> None:
        """Add a notification alert to drawer."""
        if change["new"]:
            if self.alert_badge not in self.children:
                new_children = self.children[:]
                new_children.append(self.alert_badge)
                self.children = new_children
        else:
            self.remove_notif()

        return

    def remove_notif(self) -> None:
        """Remove notification alert."""
        if self.alert_badge in self.children:
            new_children = self.children[:]
            new_children.remove(self.alert_badge)

            self.children = new_children

        return

    def display_tile(self, tiles: List[v.Card]) -> Self:
        """
        Display the apropriate tiles when the item is clicked.

        The tile to display will be all tile in the list with the mount_id as the current object.

        Args:
            tiles: the list of all the available tiles in the app
        """
        self.on_event("click", partial(self._on_click, tiles=tiles))

        return self

    def _on_click(self, tiles: List[v.Card], *args) -> Self:

        for tile in tiles:
            if self._metadata["card_id"] == tile._metadata["mount_id"]:
                tile.show()
            else:
                tile.hide()

        # trigger the resize event
        self.rt.resize()

        # change the current item status
        self.input_value = True

        # Remove notification
        self.remove_notif()

        return self


class NavDrawer(v.NavigationDrawer, SepalWidget):

    items: List[DrawerItem] = []
    "the list of all the drawerItem to display in the drawer"

    def __init__(
        self,
        items: List[DrawerItem] = [],
        code: str = "",
        wiki: str = "",
        issue: str = "",
        **kwargs,
    ) -> None:
        """
        Custom NavDrawer using the different DrawerItems of the user.

        The drawer can include links to the github page of the project for wiki, bugs and repository.

        Args:
            items: the list of all the drawerItem to display in the drawer. This items should pilote the different tile visibility
            code: the absolute link to the source code
            wiki: the absolute link the the wiki page
            issue: the absolute link to the issue tracker
            kwargs (optional) any parameter from a v.NavigationDrawer. If set, 'app' and 'children' will be overwritten.
        """
        self.items = items

        code_link = []
        if code:
            item_code = DrawerItem(
                ms.widgets.navdrawer.code, icon="fa-regular fa-file-code", href=code
            )
            code_link.append(item_code)
        if wiki:
            item_wiki = DrawerItem(
                ms.widgets.navdrawer.wiki, icon="fa-solid fa-book-open", href=wiki
            )
            code_link.append(item_wiki)
        if issue:
            item_bug = DrawerItem(
                ms.widgets.navdrawer.bug, icon="fa-solid fa-bug", href=issue
            )
            code_link.append(item_bug)

        children = [
            v.List(dense=True, children=self.items),
            v.Divider(),
            v.List(dense=True, children=code_link),
        ]

        # set default parameters
        kwargs.setdefault("v_model", True)
        kwargs["app"] = True
        kwargs.setdefault("color", color.darker)
        kwargs["children"] = children

        # call the constructor
        super().__init__(**kwargs)

        # bind the javascripts behavior
        for i in self.items:
            i.observe(self._on_item_click, "input_value")

    def display_drawer(self, toggleButton: v.Btn) -> Self:
        """
        Bind the drawer to the app toggleButton.

        Args:
            toggleButton: the button that activate the drawer
        """
        toggleButton.on_event("click", self._on_drawer_click)

        return self

    def _on_drawer_click(self, *args) -> Self:
        """
        Toggle the drawer visibility.
        """
        self.v_model = not self.v_model

        return self

    def _on_item_click(self, change: dict) -> Self:
        """
        Deactivate all the other items when on of the is activated.
        """
        if change["new"] is False:
            return self

        # reset all others states
        for i in self.items:
            if i != change["owner"]:
                i.input_value = False

        return self


class Footer(v.Footer, SepalWidget):
    def __init__(self, text: str = "", **kwargs) -> None:
        """
        Custom Footer with cuzomizable text.

        Not yet capable of displaying logos.

        Args:
            text: the text to display in the future
            kwargs (optional): any parameter from a v.Footer. If set ['app', 'children'] will be overwritten.
        """
        text = text if text != "" else "SEPAL \u00A9 {}".format(datetime.today().year)

        # set default parameters
        kwargs.setdefault("color", color.main)
        kwargs.setdefault("class_", "white--text")
        kwargs["app"] = True
        kwargs["children"] = [text]

        # call the constructor
        super().__init__(**kwargs)


class App(v.App, SepalWidget):

    tiles: List[v.Card] = []
    "the tiles of the app"

    appBar: Optional[AppBar] = None
    "the AppBar of the application"

    footer: Optional[Footer] = None
    "the footer of the application"

    navDrawer: Optional[NavDrawer] = None
    "the navdrawer of the application"

    content: Optional[v.Content] = None
    "the tiles organized in a fluid container"

    def __init__(
        self,
        tiles: List[v.Card] = [],
        appBar: Optional[AppBar] = None,
        footer: Optional[Footer] = None,
        navDrawer: Optional[NavDrawer] = None,
        translator: Optional[Translator] = None,
        **kwargs,
    ) -> None:
        """
        Custom App display with the tiles created by the user using the sepal color framework.

        Display false appBar if not filled. Navdrawer is fully optionnal.
        The drawerItem will be linked to the app tile and they will be able to control their display
        If the navdrawer exist, it will be linked to the appbar togglebtn.

        Args:
            tiles: the tiles of the app
            appBar: the appBar of the application
            footer: the footer of the application
            navDrawer: the navdrawer of the application
            translator: the translator of the app to display language informations
            kwargs (optional) any parameter from a v.App. If set, 'children' will be overwritten.
        """
        self.tiles = tiles

        app_children = []

        # create a false appBar if necessary
        if appBar is None:
            appBar = AppBar(translator=translator)
        self.appBar = appBar
        app_children.append(self.appBar)

        # add the navDrawer if existing
        self.navDrawer = None
        if navDrawer is not None:
            # bind app tile list to the navdrawer
            for di in navDrawer.items:
                di.display_tile(tiles)

            # link it with the appbar
            navDrawer.display_drawer(self.appBar.toggle_button)

            # add the drawers to the children
            self.navDrawer = navDrawer
            app_children.append(self.navDrawer)
        else:
            # remove the toggle button from the navbar
            self.appBar.toggle_button.hide()

        # add the content of the app
        self.content = v.Content(children=[v.Container(fluid=True, children=tiles)])
        app_children.append(self.content)

        # add the footer if existing
        if footer is not None:
            self.footer = footer
            app_children.append(self.footer)

        # create a negative overlay to force the background color
        bg = v.Overlay(color=color.bg, opacity=1, style_="transition:unset", z_index=-1)

        # set default parameters
        kwargs.setdefault("v_model", None)
        kwargs["children"] = [bg, *app_children]

        # call the constructor
        super().__init__(**kwargs)

        # display a warning if the set language cannot be reached
        if translator is not None:
            if translator._match is False:
                msg = ms.locale.fallback.format(
                    translator._targeted, translator._target
                )
                self.add_banner(msg, type_="error")

        # add js event
        self.appBar.locale.observe(self._locale_info, "value")
        self.appBar.theme.observe(self._theme_info, "v_model")

    def show_tile(self, name: str) -> Self:
        """
        Select the tile to display when the app is launched.

        Args:
            name: the mount-id of the tile(s) to display
        """
        # show the tile
        for tile in self.tiles:
            if name == tile._metadata["mount_id"]:
                tile.show()
            else:
                tile.hide()

        # activate the drawerItem
        if self.navDrawer:
            items = (i for i in self.navDrawer.items if i._metadata is not None)
            for i in items:
                if name == i._metadata["card_id"]:
                    i.input_value = True

        return self

    @versionadded(version="2.4.1", reason="New end user interaction method")
    @versionchanged(version="2.7.1", reason="new id\_ and persistent parameters")
    def add_banner(
        self,
        msg: str = "",
        type_: str = "info",
        id_=None,
        persistent: bool = True,
        **kwargs,
    ) -> Self:
        r"""
        Display a snackbar object on top of the app.

        Used to communicate development information to end user (release date, known issues, beta version). The alert is dissmisable and prominent.

        Args:
            msg: Message to display in application banner. default to nothing
            type\_: Used to display an appropiate banner color. fallback to "info".
            id_: unique banner identificator.
            persistent: Whether to close automatically based on the lenght of message (False) or make it indefinitely open (True). Overridden if timeout duration is set.
            **kwargs: any arguments of the sw.Banner constructor. if set, 'children' will be overwritten.
        """
        # the Banner was previously an Alert. for compatibility we accept the type parameter
        type_ = kwargs.pop("type", type_)

        # the banner will be piled up from the first to the latest.
        # only the first one is shown
        # dismissed banner are remove from the children

        # extract the banner from the app children
        children, banner_list = [], []
        for e in self.content.children.copy():
            dst = banner_list if isinstance(e, Banner) else children
            dst.append(e)

        # only set viz to true if it's the first one
        viz = False if len(banner_list) > 0 else True

        # create the baner and interactions
        w_bnr = Banner(msg, type_, id_, persistent, viz=viz, **kwargs)
        banner_list += [w_bnr]

        # display the number of banner in queue
        banner_list[0].set_btn(len(banner_list) - 1)

        # place everything back in the app chldren list
        self.content.children = banner_list + children

        # add interaction at the end
        w_bnr.observe(self._remove_banner, "v_model")

        return self

    def _locale_info(self, change: dict) -> None:
        """display information about the locale change."""
        if change["new"] != "":
            msg = ms.locale.change.format(change["new"])
            self.add_banner(msg)

        return

    def _theme_info(self, change: dict) -> None:
        """display information about the theme change."""
        if change["new"] != "":
            msg = ms.theme.change.format(change["new"])
            self.add_banner(msg)

        return

    def _remove_banner(self, change: dict) -> None:
        """
        remove banner and adapt display of the others.

        Adapt the banner display so that the first one is the only one shown displaying the number of other banner in the queue
        """
        if change["new"] is False:

            # extract the banner from the app children
            children, banner_list = [], []
            for e in self.content.children.copy():
                dst = banner_list if isinstance(e, Banner) else children
                dst.append(e)

            # remove the banner from the list
            banner_list.remove(change["owner"])

            # change the visibility of the widgets
            [setattr(b, "viz", i == 0) for i, b in enumerate(banner_list)]

            # set the btn of the the first element if possible
            len(banner_list) == 0 or banner_list[0].set_btn(len(banner_list) - 1)

            # place everything back in the app chldren list
            self.content.children = banner_list + children

        return



"""
The map displayed in the map application.
"""

from ipyleaflet import WidgetControl

from sepal_ui import mapping as sm
from sepal_ui import sepalwidgets as sw


class MapTile(sw.Tile):
    def __init__(self):
        """
        Specific Map integrating all the widget components.

        Use this map to gather all your widget and place them on it. It will reduce the amount of work to perform in the notebook
        """
        # create a map
        self.m = sm.SepalMap(zoom=3)  # to be visible on 4k screens
        self.m.add_control(
            sm.FullScreenControl(
                self.m, fullscreen=True, fullapp=True, position="topright"
            )
        )

        # create the tile
        super().__init__("map_tile", "", [self.m])

    def set_code(self, link):
        """Add the code link btn to the map."""
        btn = sm.MapBtn("fa-solid fa-code", href=link, target="_blank")
        control = WidgetControl(widget=btn, position="bottomleft")
        self.m.add(control)

        return

    def set_wiki(self, link):
        """Add the wiki link btn to the map."""
        btn = sm.MapBtn("fa-solid fa-book-open", href=link, target="_blank")
        control = WidgetControl(widget=btn, position="bottomleft")
        self.m.add(control)

        return

    def set_issue(self, link):
        """Add the code link btn to the map."""
        btn = sm.MapBtn("fa-solid fa-bug", href=link, target="_blank")
        control = WidgetControl(widget=btn, position="bottomleft")
        self.m.add(control)

        return



"""
Widgets used to build the ``PLanetView`` interface.
"""

from datetime import datetime, timezone
from typing import List, Optional

import ipyvuetify as v
from typing_extensions import Self

import sepal_ui.sepalwidgets as sw
from sepal_ui.planetapi.planet_model import PlanetModel

# key_name: [label, [non_active_color, active_color]]
BTNS: dict = {
    "nicfi": ["NICFI", ["menu", "success"]],
    "others": ["Others", ["menu", "success"]],
}


class InfoView(sw.ExpansionPanels):

    model: Optional[PlanetModel] = None
    "Backend model to manipulate interface actions"

    def __init__(self, model: PlanetModel, **kwargs) -> None:
        """
        Card to validate subscription.

        Custom optinal card to be displayed within the planet view to validate the available
        subscriptions from the log-in credentials and show the info related with them, such
        as the quotas and remaining time of activation.

        Args:
            model: the planetModel associated with the display
        """
        self.model = model
        self.v_model = 1
        self.current = None
        self.readonly = True

        super().__init__(**kwargs)

        subs_btn = [
            sw.Chip(
                children=[BTNS[label][0]],
                disabled=True,
                color=BTNS[label][1][0],
                x_small=True,
                class_="mr-2",
                attributes={"id": label},
                link=True,
                label=True,
            )
            for label in BTNS.keys()
        ]

        self.info_card = InfoCard().hide()

        self.children = [
            v.ExpansionPanel(
                children=[
                    v.ExpansionPanelHeader(
                        hide_actions=True, children=[v.Flex(children=subs_btn)]
                    ),
                    v.ExpansionPanelContent(v_model=1, children=[self.info_card]),
                ]
            )
        ]

        [chip.on_event("click", self.open_info) for chip in subs_btn]

        self.model.observe(self._toggle_btns, "subscriptions")

    def open_info(self, widget: v.VuetifyWidget, *args) -> None:
        """
        Shrink or srhunk the content of the expansion panel.

        It automatically sends a request to build the data.

        Args:
            widget: the widget to expand
        """
        is_current = self.current == widget.attributes["id"]
        self.v_model = (not self.v_model) * 1 if is_current else 0
        self.current = widget.attributes["id"]

        subs_group = self.model.subscriptions[widget.attributes["id"]]

        self.info_card.update(subs_group).show()

        return

    def _turn_btn(self, btn_id: str, state: bool) -> None:
        """
        Update the status of the given button.

        Args:
            btn_id: the id of the btn object
            state: the state to apply to the btn
        """
        btn = self.get_children(attr="id", value=btn_id)[0]
        btn.disabled = not state
        btn.color = BTNS[btn_id][1][state]

        return

    def _toggle_btns(self, change: dict) -> None:
        """
        Toggle the status of the btns.
        """
        if not change["new"]:
            self.v_model = 1
            [self._turn_btn(btn_id, False) for btn_id in BTNS.keys()]
            return

        for plan_type in BTNS.keys():
            for subs in self.model.subscriptions[plan_type]:
                plan = subs.get("plan")
                # Turn on if at least one of them is True
                state = True if plan.get("state") else False
                self._turn_btn(plan_type, state)
                if state:
                    break

        return


class InfoCard(sw.Layout):
    def __init__(self) -> None:
        """
        Information card that will display the subscription data.
        """
        self.style_ = "max-height: 240px; overflow: auto"
        self.class_ = "d-block"

        super().__init__()

        self.children = [v.CardText(children=[])]

    def _make_content(self, sub: dict) -> List[v.VuetifyWidget]:
        """
        Creates individual subscription card from a subscription list.

        Args:
            sub: the subscriptions plan from a defined category (e.g. "nicfi")

        Returns:
            the children content of a category
        """
        title = sub["plan"]["name"].replace("_", " ")
        state = sub["plan"]["state"]

        # Create an individual State icon for all the elements, it has to be
        # independant
        w_state = sw.StateIcon(
            states={
                "non_active": ["Non active", "error"],
                "active": ["Active", "success"],
            }
        )
        w_state.values = state

        w_title = sw.CardTitle(children=[title, v.Spacer(), w_state])
        w_subtitle = v.CardSubtitle(children=[state.capitalize()])

        from_ = datetime.fromisoformat(sub["active_from"])
        to = sub["active_to"] is not None and datetime.fromisoformat(sub["active_to"])
        now = datetime.now(timezone.utc)
        days_left = "∞" if not to else (to - now).days

        info_dict = {
            "from": ["From:", from_.strftime("%Y/%m/%d")],
            "to": ["Until:", "∞" if not to else to.strftime("%Y/%m/%d")],
            "days_left": ["Days left:", f"{days_left}"],
        }

        content = [
            (
                v.Flex(
                    class_="d-block",
                    children=[
                        v.Html(tag="strong", children=[values[0]]),
                        v.Html(tag="div", children=[values[1]]),
                    ],
                ),
                v.Divider(vertical=True, class_="mx-4"),
            )
            for values in info_dict.values()
        ]
        # Flat the nested elements and remove the last divider
        content = [e for row in content for e in row][:-1]

        return [
            w_title,
            w_subtitle,
            v.Layout(class_="d-flex flex-wrap", children=content),
            v.Divider(class_="my-2"),
        ]

    def update(self, subs_group: List[dict]) -> Self:
        """
        Extract the info from the subscription and set it in the card.

        Args:
            subs_group: list of subscriptions belonging to the same category ('nicfi', 'others')
        """
        content = [
            v.Card(class_="pa-2", children=self._make_content(sub))
            for sub in subs_group
        ]

        self.children = content

        return self



"""
Widget control providing zoom options for the end user.
"""

from typing import Optional, Union

import ee
import ipyvuetify as v
from ipyleaflet import Map
from shapely import geometry as sg

from sepal_ui import sepalwidgets as sw
from sepal_ui.mapping.menu_control import MenuControl
from sepal_ui.scripts import decorator as sd


class AoiControl(MenuControl):

    m: Optional[Map] = None
    "the map on which he AoiControl is displayed to interact with itszoom and center"

    aoi_list: Optional[sw.ListItemGroup] = None
    "the list of ListItem used in the menu. each one as the name provided by the user and the value of the bounds"

    aoi_bounds: dict = {}
    "the list of bounds to fit on. using the aoi names as keys"

    def __init__(self, m: Map, **kwargs) -> None:
        """
        Widget control providing zoom options for the end user.

        The developer can add as many gemetries to the widget and the user will simply have to click on them to move to the appropriate AOI.

        Args:
            m: the map on which he AoiControl is displayed to interact with itszoom and center
        """
        # init the aoi data list
        self.aoi_bounds = {}

        # set some default parameters
        kwargs.setdefault("position", "topright")
        kwargs["m"] = m

        # create a list
        self.aoi_list = sw.ListItemGroup(children=[], v_model="")

        # create the widget
        super().__init__("fa-solid fa-search-location", self.aoi_list, **kwargs)

        # change a bit the behavior of the control
        self.menu.open_on_hover = True
        self.menu.open_on_click = False
        self.menu.cole_on_content_click = True

        # set the size of the card to 0 so that the list controls the widget display
        self.set_size(
            min_width="200px", max_width="200px", min_height=None, max_height="300px"
        )

        # add js behaviours
        self.menu.v_slots[0]["children"].on_event("click", self.click_btn)

    def click_btn(self, *args) -> None:
        """
        Zoom to the total area of all AOIs in :code:`self.aoi_bounds`. Use the whole world if empty.
        """
        # set the bounds to the world if the list is empty
        if len(self.aoi_bounds) == 0:
            self.m.center = [0, 0]
            self.m.zoom = 2.0
        else:
            minx = min([i[0] for i in self.aoi_bounds.values()])
            miny = min([i[1] for i in self.aoi_bounds.values()])
            maxx = max([i[2] for i in self.aoi_bounds.values()])
            maxy = max([i[3] for i in self.aoi_bounds.values()])
            bounds = (minx, miny, maxx, maxy)
            self.m.zoom_bounds(bounds)

        return

    @sd.need_ee
    def add_aoi(
        self, name: str, item: Union[sg.base.BaseGeometry, ee.ComputedObject]
    ) -> None:
        """
        Add an AOI to the list and refresh the list displayed. the AOI will be composed of a name and the bounds of the provided item.

        Args:
            name: the name of the AOI
            item: the item to use to compute the bounds. It need to be a shapely geometry or an ee object.
        """
        if isinstance(item, ee.ComputedObject):
            # extract bounds from ee_object
            ee_geometry = item if isinstance(item, ee.Geometry) else item.geometry()
            bl, br, tr, tl, _ = ee_geometry.bounds().coordinates().get(0).getInfo()
            bounds = (*bl, *tr)

        elif isinstance(item, sg.base.BaseGeometry):
            bounds = item.bounds

        else:
            raise ValueError("Cannot extract bounds from the input")

        self.aoi_bounds[name] = bounds

        # update the list
        self.update_list()

        return

    def remove_aoi(self, name: str) -> None:
        """
        Remove an item from the :code:`self.aoi_bounds` dict and from the ListItem. It will raise a KeyError if the name cannot be found.

        Args:
            name: the name of the aoi to remove
        """
        # remove the value from the list
        self.aoi_bounds.pop(name)

        # update the list
        self.update_list()

        return

    def zoom(self, widget: v.VuetifyWidget, *args) -> None:
        """
        Zoom on the specified bounds.

        Args:
            widget: the clicked widget containing the bounds
        """
        # the widget store the bounding box in value
        # a tuple of the bounds of the geometry (minx, miny, maxx, maxy)
        self.m.zoom_bounds(widget.value)

        return

    def update_list(self) -> None:
        """
        Update the ListItem children of the object based on the content of :code:`self.aoi_bounds`.
        """
        children = []

        for name, bounds in self.aoi_bounds.items():

            text = sw.ListItemContent(children=[sw.ListItemTitle(children=[name])])
            item = sw.ListItem(dense=True, value=bounds, children=[text])
            children.append(item)
            item.on_event("click", self.zoom)

        self.aoi_list.children = children

        return



"""
Custom Card widget to create Tiles in application.

Gather the customized ``ipyvuetifyWidgets`` used to create Tiles in the application panel framework.
All the content of this modules is included in the parent ``sepal_ui.sepalwidgets`` package. So it can be imported directly from there.

Example:
    .. jupyter-execute::
    
        from sepal_ui import sepalwidgets as sw
        
        sw.TileDisclaimer()
"""

from pathlib import Path
from typing import List, Optional, Union

import ipyvuetify as v
from typing_extensions import Self

from sepal_ui.message import ms
from sepal_ui.scripts import utils as su
from sepal_ui.sepalwidgets.sepalwidget import SepalWidget
from sepal_ui.sepalwidgets.widget import Markdown

__all__ = ["Tile", "TileAbout", "TileDisclaimer"]


class Tile(v.Layout, SepalWidget):

    btn: Optional[v.Btn] = None
    "the process btn"

    alert: Optional[v.Alert] = None
    "the alert to display process informations to the end user"

    title: Optional[v.Html] = None
    "the title of the Tile"

    def __init__(
        self,
        id_: str,
        title: str,
        inputs: list = [""],
        btn: Union[v.Btn, None] = None,
        alert: Union[v.Alert, None] = None,
        **kwargs,
    ) -> None:
        """
        Custom Layout widget for the sepal UI framework.

        It is an helper to build a consistent tiling system. Tile objects are indeed compatible with the other classes from sepal_ui.

        Args:
            id_: the tile id that will be written in its mount_id _metadata attribute
            title: the title of the Tile
            inputs: the list of widget to display inside the tile
            btn: the process btn
            alert: the alert to display process informations to the end user
            kwargs: any parameter from a v.Layout. if set, 'children' and '_metadata' will be overwritten.
        """
        self.btn = btn
        inputs += [btn] if btn else []

        self.alert = alert
        inputs += [alert] if alert else []

        self.title = v.Html(xs12=True, tag="h2", children=[title])

        content = [v.Flex(xs12=True, children=[widget]) for widget in inputs]

        card = v.Card(
            class_="pa-5", raised=True, xs12=True, children=[self.title] + content
        )

        # set some default parameters
        kwargs["_metadata"] = {"mount_id": id_}
        kwargs.setdefault("row", True)
        kwargs.setdefault("align_center", True)
        kwargs.setdefault("class_", "ma-5 d-inline")
        kwargs["children"] = [card]

        # call the constructor
        super().__init__(**kwargs)

    def nest(self) -> Self:
        """
        Prepare the tile to be used as a nested component in a tile.

        the elevation will be set to 0 and the title remove from children.
        The mount_id will also be changed to nested.
        """
        # remove id
        self._metadata["mount_id"] = "nested_tile"

        # remove elevation
        self.children[0].elevation = False

        # remove title
        self.set_title()

        return self

    def set_content(self, inputs: list) -> Self:
        """
        Replace the current content of the tile with the provided inputs.

        It will keep the output and btn widget if existing.

        Args:
            the list of widget to display inside the tile
        """
        # create the widgets
        content = [v.Flex(xs12=True, children=[widget]) for widget in inputs]

        # add the output (if existing)
        if self.title:
            content = [self.title] + content

        if self.alert:
            content = content + [self.alert]

        if self.btn:
            content = content + [self.btn]

        self.children[0].children = content

        return self

    def set_title(self, title: str = "") -> Self:
        """
        Replace the current title and activate it.

        If no title is provided, the title is removed from the tile content.

        Args:
            title: the new title of the object
        """
        # set the title text
        self.title.children = [title]

        # add the title if it's deactivated
        if title and self.title not in self.children[0].children:
            self.children[0].children = [self.title] + self.children[0].children.copy()

        # remove it if it's deactivated
        elif self.title in self.children[0].children and not title:
            # it's the first one
            self.children[0].children = self.children[0].children.copy()[1:]

        return self

    def get_title(self) -> str:
        """
        Return the current title of the tile.

        Returns:
            the title
        """
        return self.title.children[0]

    def toggle_inputs(
        self, fields_2_show: List[v.VuetifyWidget], fields: List[v.VuetifyWidget]
    ) -> Self:
        """
        Display only the widgets that are part of the input_list.

        The widget_list is the list of all the widgets of the tile.

        Args:
            fields_2_show: the list of input to be display
            fields: the list of the tile widget
        """
        for field in fields:
            if field in fields_2_show:
                su.show_component(field)
            else:
                su.hide_component(field)

        return self

    def get_id(self) -> str:
        """
        Get the mount_id value.

        Returns:
            the moun_id value from _metadata dict
        """
        return self._metadata["mount_id"]


class TileAbout(Tile):
    def __init__(self, pathname: Union[str, Path], **kwargs) -> None:
        """
        Create an about tile using a .md file.

        This tile will have the "about_widget" id and "About" title.

        Args:
            pathname: the path to the .md file
        """
        content = Markdown(Path(pathname).read_text())
        super().__init__("about_tile", "About", inputs=[content])


class TileDisclaimer(Tile):
    def __init__(self) -> None:
        """
        Create an about tile.

        This tile will have the "about_widget" id and "Disclaimer" title.
        """
        # create the tile content on the fly
        disclaimer = "  \n".join(ms.disclaimer.p)
        disclaimer += "  \n"
        disclaimer += '<div style="inline-block">'

        # add the logo (href, src, alt)
        logo_list = [
            ("http://www.fao.org/home/en/", "fao.png", "fao_logo"),
            ("http://www.openforis.org", "open-foris.png", "openforis_logo"),
            ("https://sepal.io", "sepal.png", "sepal_logo"),
        ]
        theme = "dark" if v.theme.dark is True else "light"
        url = f"https://raw.githubusercontent.com/12rambau/sepal_ui/master/sepal_ui/frontend/images/{theme}"
        for href, src, alt in logo_list:
            disclaimer += f'<a href="{href}"><img src="{url}/{src}" alt="{alt}" height="100" class="ma-3"/></a>'

        # close the file
        disclaimer += "</div>"

        content = Markdown(disclaimer)

        super().__init__("about_tile", "Disclaimer", inputs=[content])



"""
All modules used to create an Planet interface.

Package to access all the widget, model and tiles to create an Planet interface in sepal-ui.

Every module content can be called directly from the package.

Example:
    .. jupyter-execute::

        from sepal_ui import planetapi

        planetapi.PlanetView()
"""

from .planet_model import *
from .planet_view import *
